__author__ = "sibirrer"

import numpy as np
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["SIS"]


class SIS(LensProfileBase):
    """This class contains the function and the derivatives of the Singular Isothermal
    Sphere.

    .. math::
        \\kappa(x, y) = \\frac{1}{2} \\left(\\frac{\\theta_{E}}{\\sqrt{x^2 + y^2}} \\right)

    with :math:`\\theta_{E}` is the Einstein radius,
    """

    param_names = ["theta_E", "center_x", "center_y"]
    lower_limit_default = {"theta_E": 0, "center_x": -100, "center_y": -100}
    upper_limit_default = {"theta_E": 100, "center_x": 100, "center_y": 100}

    def function(self, x, y, theta_E, center_x=0, center_y=0):
        x_shift = x - center_x
        y_shift = y - center_y
        f_ = theta_E * np.sqrt(x_shift * x_shift + y_shift * y_shift)
        return f_

    def derivatives(self, x, y, theta_E, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function."""
        x_shift = x - center_x
        y_shift = y - center_y
        R = np.sqrt(x_shift * x_shift + y_shift * y_shift)
        if isinstance(R, int) or isinstance(R, float):
            a = theta_E / max(0.000001, R)
        else:
            a = np.empty_like(R)
            r = R[R > 0]  # in the SIS regime
            a[R == 0] = 0
            a[R > 0] = theta_E / r
        f_x = a * x_shift
        f_y = a * y_shift
        return f_x, f_y

    def hessian(self, x, y, theta_E, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^2/dxdy, d^2/dydx,
        d^f/dy^2."""
        x_shift = x - center_x
        y_shift = y - center_y
        R = (x_shift * x_shift + y_shift * y_shift) ** (3.0 / 2)
        if isinstance(R, int) or isinstance(R, float):
            prefac = theta_E / max(0.000001, R)
        else:
            prefac = np.empty_like(R)
            r = R[R > 0]  # in the SIS regime
            prefac[R == 0] = 0.0
            prefac[R > 0] = theta_E / r

        f_xx = y_shift * y_shift * prefac
        f_yy = x_shift * x_shift * prefac
        f_xy = -x_shift * y_shift * prefac
        return f_xx, f_xy, f_xy, f_yy

    @staticmethod
    def rho2theta(rho0):
        """Converts 3d density into 2d projected density parameter :param rho0:

        :return:
        """
        theta_E = np.pi * 2 * rho0
        return theta_E

    @staticmethod
    def theta2rho(theta_E):
        """Converts projected density parameter (in units of deflection) into 3d density
        parameter :param theta_E: Einstein radius :return:"""
        fac1 = np.pi * 2
        rho0 = theta_E / fac1
        return rho0

    @staticmethod
    def mass_3d(r, rho0):
        """Mass enclosed a 3d sphere or radius r :param r: radius in angular units
        :param rho0: density at angle=1 :return: mass in angular units."""
        mass_3d = 4 * np.pi * rho0 * r
        return mass_3d

    def mass_3d_lens(self, r, theta_E):
        """Mass enclosed a 3d sphere or radius r given a lens parameterization with
        angular units.

        :param r: radius in angular units
        :param theta_E: Einstein radius
        :return: mass in angular units
        """
        rho0 = self.theta2rho(theta_E)
        return self.mass_3d(r, rho0)

    @staticmethod
    def mass_2d(r, rho0):
        """Mass enclosed projected 2d sphere of radius r :param r:

        :param rho0:
        :return:
        """
        alpha = 2 * rho0 * np.pi**2
        mass_2d = alpha * r
        return mass_2d

    def mass_2d_lens(self, r, theta_E):
        """

        :param r: radius
        :param theta_E: Einstein radius
        :return: mass within a radius in projection
        """
        rho0 = self.theta2rho(theta_E)
        return self.mass_2d(r, rho0)

    def grav_pot(self, x, y, rho0, center_x=0, center_y=0):
        """Gravitational potential (modulo 4 pi G and rho0 in appropriate units) :param
        x:

        :param y:
        :param rho0:
        :param center_x:
        :param center_y:
        :return:
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        mass_3d = self.mass_3d(r, rho0)
        pot = mass_3d / r
        return pot

    @staticmethod
    def density(r, rho0):
        """Computes the density :param r: radius in angles :param rho0: density at
        angle=1 :return: density at r."""
        rho = rho0 / r**2
        return rho

    def density_lens(self, r, theta_E):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in projected in units of angles (i.e. arc seconds) results in the
        convergence quantity.

        :param r: 3d radius
        :param theta_E: Einstein radius
        :return: density(r)
        """
        rho0 = self.theta2rho(theta_E)
        return self.density(r, rho0)

    @staticmethod
    def density_2d(x, y, rho0, center_x=0, center_y=0):
        """Projected density :param x:

        :param y:
        :param rho0:
        :param center_x:
        :param center_y:
        :return:
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        sigma = np.pi * rho0 / r
        return sigma



__author__ = "lynevdv"


from lenstronomy.LensModel.Profiles.multipole import Multipole, EllipticalMultipole
from lenstronomy.Util import util

import numpy as np
import pytest
import numpy.testing as npt


class TestMultipole(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.Multipole = Multipole()

    def test_function(self):
        x = 1
        y = 2
        m = 4
        a_m = 0.05
        phi_m = 25 * np.pi / 180.0
        values = self.Multipole.function(x, y, m, a_m, phi_m)
        npt.assert_almost_equal(values, 0.006684307, decimal=6)
        x = np.array([0])
        y = np.array([0])
        values = self.Multipole.function(x, y, m, a_m, phi_m)
        assert values[0] == 0

        x = np.array([2, 3, 4])
        y = np.array([1, 1, 1])
        values = self.Multipole.function(x, y, m, a_m, phi_m)
        npt.assert_almost_equal(values[0], -0.007409114, decimal=6)
        npt.assert_almost_equal(values[1], -0.009453038, decimal=6)
        npt.assert_almost_equal(values[2], -0.009910505, decimal=6)

    def test_derivatives(self):
        x = 1
        y = 2
        m = 4
        a_m = 0.05
        phi_m = 25 * np.pi / 180.0
        f_x, f_y = self.Multipole.derivatives(x, y, m, a_m, phi_m)
        npt.assert_almost_equal(f_x, -0.003939644, decimal=6)
        npt.assert_almost_equal(f_y, 0.005311976, decimal=6)

        x = np.array([1])
        y = np.array([2])
        f_x, f_y = self.Multipole.derivatives(x, y, m, a_m, phi_m)
        npt.assert_almost_equal(f_x[0], -0.003939644, decimal=6)
        npt.assert_almost_equal(f_y[0], 0.005311976, decimal=6)

        x = np.array([2, 3, 1])
        y = np.array([1, 1, 4])
        f_x, f_y = self.Multipole.derivatives(x, y, m, a_m, phi_m)
        npt.assert_almost_equal(f_x[0], -0.003613858, decimal=6)
        npt.assert_almost_equal(f_x[1], -0.000970385, decimal=6)
        npt.assert_almost_equal(f_x[2], 0.005970704, decimal=6)
        npt.assert_almost_equal(f_y[0], -0.000181398, decimal=6)
        npt.assert_almost_equal(f_y[1], -0.006541883, decimal=6)
        npt.assert_almost_equal(f_y[2], 0.001649720, decimal=6)

    def test_hessian(self):
        x = 1
        y = 2
        m = 4
        a_m = 0.05
        phi_m = 25 * np.pi / 180.0
        f_xx, f_xy, f_yx, f_yy = self.Multipole.hessian(x, y, m, a_m, phi_m)
        npt.assert_almost_equal(f_xx, -0.016042338, decimal=6)
        npt.assert_almost_equal(f_yy, -0.004010584, decimal=6)
        npt.assert_almost_equal(f_xy, 0.008021169, decimal=6)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)
        x = np.array([1])
        y = np.array([2])
        f_xx, f_xy, f_yx, f_yy = self.Multipole.hessian(x, y, m, a_m, phi_m)
        npt.assert_almost_equal(f_xx[0], -0.016042338, decimal=6)
        npt.assert_almost_equal(f_yy[0], -0.004010584, decimal=6)
        npt.assert_almost_equal(f_xy[0], 0.008021169, decimal=6)
        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        values = self.Multipole.hessian(x, y, m, a_m, phi_m)
        npt.assert_almost_equal(values[0][0], -0.016042338, decimal=6)
        npt.assert_almost_equal(values[3][0], -0.004010584, decimal=6)
        npt.assert_almost_equal(values[1][0], 0.008021169, decimal=6)
        npt.assert_almost_equal(values[0][1], 0.001417956, decimal=6)
        npt.assert_almost_equal(values[3][1], 0.012761602, decimal=6)
        npt.assert_almost_equal(values[1][1], -0.004253867, decimal=6)


class TestEllipticalMultipole(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.Multipole = EllipticalMultipole()
        self.SphericalMultipole = Multipole()
        self.x, self.y = util.make_grid(numPix=10, deltapix=0.2)

    def test_function(self):
        x = 1
        y = 2
        q = 0.5
        m = 4
        a_m = 0.05
        phi_m = 25 * np.pi / 180.0
        values = self.Multipole.function(x, y, m, a_m, phi_m, q)
        npt.assert_almost_equal(values, 0.017025477, decimal=6)
        x = np.array([0])
        y = np.array([0])
        values = self.Multipole.function(x, y, m, a_m, phi_m, q)
        assert values[0] == 0

        x = np.array([2, 3, 4])
        y = np.array([1, 1, 1])
        values = self.Multipole.function(x, y, m, a_m, phi_m, q)
        npt.assert_almost_equal(values[0], 0.005762831, decimal=6)
        npt.assert_almost_equal(values[1], 0.000954611, decimal=6)
        npt.assert_almost_equal(values[2], -0.002321308, decimal=6)

        # Test that the limit q-> 1 is consistent with the spherical multipoles
        function_ell_limit = self.Multipole.function(
            self.x, self.y, m=4, a_m=a_m, phi_m=np.pi / 24, q=0.99989
        )
        function_sph = self.SphericalMultipole.function(
            self.x, self.y, m=4, a_m=a_m, phi_m=np.pi / 24
        )
        npt.assert_allclose(function_ell_limit, function_sph, rtol=1e-5, atol=5e-5)

        function_ell_limit = self.Multipole.function(
            self.x, self.y, m=3, a_m=a_m, phi_m=np.pi / 18, q=0.99989
        )
        function_sph = self.SphericalMultipole.function(
            self.x, self.y, m=3, a_m=a_m, phi_m=np.pi / 18
        )
        npt.assert_allclose(function_ell_limit, function_sph, rtol=1e-5, atol=5e-5)

    def test_derivatives(self):
        x = 1
        y = 2
        q = 0.5
        m = 4
        a_m = 0.05
        phi_m = 25 * np.pi / 180.0
        f_x, f_y = self.Multipole.derivatives(x, y, m, a_m, phi_m, q)
        npt.assert_almost_equal(f_x, 0.009253496, decimal=6)
        npt.assert_almost_equal(f_y, 0.003885991, decimal=6)

        x = np.array([2, 3, 1])
        y = np.array([1, 1, 4])
        f_x, f_y = self.Multipole.derivatives(x, y, m, a_m, phi_m, q)
        npt.assert_almost_equal(f_x[0], -0.005559822, decimal=6)
        npt.assert_almost_equal(f_x[1], -0.003963335, decimal=6)
        npt.assert_almost_equal(f_x[2], 0.015182005, decimal=6)
        npt.assert_almost_equal(f_y[0], 0.016882475, decimal=6)
        npt.assert_almost_equal(f_y[1], 0.012844618, decimal=6)
        npt.assert_almost_equal(f_y[2], 0.001639347, decimal=6)

        # Test that the limit q-> 1 is consistent with the spherical multipoles
        alpha_x_ell_limit, alpha_y_ell_limit = self.Multipole.derivatives(
            self.x, self.y, m=4, a_m=a_m, phi_m=np.pi / 24, q=0.99989
        )
        alpha_x_sph, alpha_y_sph = self.SphericalMultipole.derivatives(
            self.x, self.y, m=4, a_m=a_m, phi_m=np.pi / 24
        )
        npt.assert_allclose(alpha_x_ell_limit, alpha_x_sph, rtol=1e-5, atol=5e-5)
        npt.assert_allclose(alpha_y_ell_limit, alpha_y_sph, rtol=1e-5, atol=5e-5)

        alpha_x_ell_limit, alpha_y_ell_limit = self.Multipole.derivatives(
            self.x, self.y, m=3, a_m=a_m, phi_m=np.pi / 18, q=0.99989
        )
        alpha_x_sph, alpha_y_sph = self.SphericalMultipole.derivatives(
            self.x, self.y, m=3, a_m=a_m, phi_m=np.pi / 18
        )
        npt.assert_allclose(alpha_x_ell_limit, alpha_x_sph, rtol=1e-5, atol=5e-5)
        npt.assert_allclose(alpha_y_ell_limit, alpha_y_sph, rtol=1e-5, atol=5e-5)

    def test_hessian(self):
        x = 1
        y = 2
        q = 0.5
        m = 4
        a_m = 0.05
        phi_m = 25 * np.pi / 180.0
        f_xx, f_xy, f_yx, f_yy = self.Multipole.hessian(x, y, m, a_m, phi_m, q)
        npt.assert_almost_equal(f_xx, -0.01254782, decimal=6)
        npt.assert_almost_equal(f_yy, -0.003136955, decimal=6)
        npt.assert_almost_equal(f_xy, 0.00627391, decimal=6)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)

        x = np.array([2, 3, 4])
        y = np.array([1, 1, 1])
        values = self.Multipole.hessian(x, y, m, a_m, phi_m, q)
        npt.assert_almost_equal(values[0][0], 0.000868241, decimal=6)
        npt.assert_almost_equal(values[0][1], 0.001611182, decimal=6)
        npt.assert_almost_equal(values[0][2], 0.000924536, decimal=6)
        npt.assert_almost_equal(values[3][0], 0.003472964, decimal=6)
        npt.assert_almost_equal(values[3][1], 0.014500636, decimal=6)
        npt.assert_almost_equal(values[3][2], 0.014792568, decimal=6)
        npt.assert_almost_equal(values[1][0], -0.001736482, decimal=6)
        npt.assert_almost_equal(values[1][1], -0.004833545, decimal=6)
        npt.assert_almost_equal(values[1][2], -0.003698142, decimal=6)

        # Test that the limit q-> 1 is consistent with the spherical multipoles
        f_xx_ell_limit, f_xy_ell_limit, f_yx_ell_limit, f_yy_ell_limit = (
            self.Multipole.hessian(
                self.x, self.y, m=4, a_m=a_m, phi_m=np.pi / 24, q=0.99989
            )
        )
        f_xx_sph, f_xy_sph, f_yx_sph, f_yy_sph = self.SphericalMultipole.hessian(
            self.x, self.y, m=4, a_m=a_m, phi_m=np.pi / 24
        )
        npt.assert_allclose(f_xx_ell_limit, f_xx_sph, rtol=1e-5, atol=5e-5)
        npt.assert_allclose(f_xy_ell_limit, f_xy_sph, rtol=1e-5, atol=5e-5)
        npt.assert_allclose(f_yx_ell_limit, f_yx_sph, rtol=1e-5, atol=5e-5)
        npt.assert_allclose(f_yy_ell_limit, f_yy_sph, rtol=1e-5, atol=5e-5)

        f_xx_ell_limit, f_xy_ell_limit, _, f_yy_ell_limit = self.Multipole.hessian(
            self.x, self.y, m=3, a_m=a_m, phi_m=np.pi / 18, q=0.99989
        )
        f_xx_sph, f_xy_sph, _, f_yy_sph = self.SphericalMultipole.hessian(
            self.x, self.y, m=3, a_m=a_m, phi_m=np.pi / 18
        )
        npt.assert_allclose(f_xx_ell_limit, f_xx_sph, rtol=1e-5, atol=5e-5)
        npt.assert_allclose(f_xy_ell_limit, f_xy_sph, rtol=1e-5, atol=5e-5)
        npt.assert_allclose(f_yx_ell_limit, f_yx_sph, rtol=1e-5, atol=5e-5)
        npt.assert_allclose(f_yy_ell_limit, f_yy_sph, rtol=1e-5, atol=5e-5)


if __name__ == "__main__":
    pytest.main()



import numpy as np
from scipy.special import hyp2f1
from scipy import interpolate
from astropy.cosmology import default_cosmology

from .nfw_param import NFWParam

__all__ = ["GNFWParam"]


class GNFWParam(object):
    """Class which contains a halo model parameters dependent on cosmology for gNFW
    profile. All distances are given in physical units.

    Mass definitions are relative to 200 crit including redshift evolution. The redshift
    evolution is cosmology dependent (dark energy). The H0 dependence is propagated into
    the input and return units.
    """

    rhoc = 2.77536627e11  # critical density [h^2 M_sun Mpc^-3]

    def __init__(self, cosmo=None):
        """

        :param cosmo: astropy.cosmology instance
        :type cosmo: astropy.cosmology instance
        """
        if cosmo is None:
            cosmo = default_cosmology.get()
        self.cosmo = cosmo
        self.nfw_param = NFWParam(cosmo)

    def rhoc_z(self, z):
        """Compute the critical density of the universe at redshift z in physical units
        [h^2 M_sun Mpc^-3].

        :param z: redshift
        :type z: float
        :return: critical density of the universe at redshift z in physical units [h^2
            M_sun Mpc^-3]
        :rtype: float
        """
        return self.nfw_param.rhoc_z(z)
        # return self.rhoc*(1+z)**3

    @staticmethod
    def M200(rs, rho0, c, gamma_in):
        """Calculation of the mass enclosed r_200 for gNFW profile defined as.

        .. math::
            M_{200} = 4 \\pi \\rho_0^{3} r_{\\rm s}^{3} \frac{c^{3 - \\gamma_{\\rm in}}} {3 - \\gamma_{\\rm in}}  {}_2F_1(3 - \\gamma_{\\rm in}, 3 - \\gamma_{\\rm in}; 4 - \\gamma_{\\rm in}; -c)

        :param rs: scale radius
        :type rs: float
        :param rho0: density normalization (characteristic density) in units mass/[distance unit of rs]^3
        :type rho0: float
        :param c: concentration
        :type c: float [4,40]
        :param gamma_in: inner slope of the gNFW profile
        :type gamma_in: float
        :return: M(R_200) mass in units of rho0 * rs^3
        :rtype: float
        """
        return (
            4
            * np.pi
            * rho0
            * rs**3
            * c ** (3 - gamma_in)
            / (3 - gamma_in)
            * hyp2f1(3 - gamma_in, 3 - gamma_in, 4 - gamma_in, -c)
        )

    def r200_M(self, M, z):
        """Compute the radius R_200 crit of a halo of mass M in physical mass M/h.

        :param M: halo mass in M_sun/h
        :type M: float or numpy array
        :param z: redshift
        :type z: float
        :return: radius R_200 in physical Mpc/h
        :rtype: float or numpy array
        """
        return self.nfw_param.r200_M(M, z)

    def M_r200(self, r200, z):
        """Compute the mass M_200 of a halo of radius r_200 in physical Mpc/h.

        :param r200: r200 in physical Mpc/h
        :type r200: float
        :param z: redshift
        :type z: float
        :return: M200 in M_sun/h
        :rtype: float
        """
        return self.nfw_param.M_r200(r200, z)

    def rho0_c(self, c, z, gamma_in):
        """Computes density normalization as a function of concentration parameter.

        :param c: concentration
        :type c: float
        :param z: redshift
        :type z: float
        :param gamma_in: inner slope of the gNFW profile
        :type gamma_in: float
        :return: density normalization in h^2/Mpc^3 (physical)
        :rtype: float
        """
        return (
            200.0
            / 3
            * self.rhoc_z(z)
            * (3 - gamma_in)
            * c**gamma_in
            / hyp2f1(3 - gamma_in, 3 - gamma_in, 4 - gamma_in, -c)
        )

    def c_rho0(self, rho0, z, gamma_in):
        """Computes the concentration given density normalization rho_0 in h^2/Mpc^3
        (physical) (inverse of function rho0_c)

        :param rho0: density normalization in h^2/Mpc^3 (physical)
        :type rho0: float
        :param z: redshift
        :type z: float
        :param gamma_in: inner slope of the gNFW profile
        :type gamma_in: float
        :return: concentration parameter c
        :rtype: float
        """
        c_array = np.linspace(0.1, 30, 100)
        if not hasattr(self, "_rho0_c_gamma_in_interps"):
            gamma_in_array = np.linspace(0.1, 2.99, 100)
            self._rho0_c_gamma_in_interps = []

            for i in range(len(c_array)):
                rho0_array = self.rho0_c(c_array[i], z, gamma_in_array)

                self._rho0_c_gamma_in_interps.append(
                    interpolate.InterpolatedUnivariateSpline(
                        gamma_in_array, rho0_array, w=None, bbox=[None, None], k=3
                    )
                )

        rho0_interp = [interp(gamma_in) for interp in self._rho0_c_gamma_in_interps]

        c_rho0_interp = interpolate.InterpolatedUnivariateSpline(
            rho0_interp, c_array, w=None, bbox=[None, None], k=3
        )

        return c_rho0_interp(rho0)

    def c_M_z(self, M, z):
        """
        Fitting function of http://moriond.in2p3.fr/J08/proceedings/duffy.pdf for the mass and redshift dependence of
        the concentration parameter. Here, assuming the NFW M-c relation for the gNFW profile.

        :param M: halo mass in M_sun/h
        :type M: float or numpy array
        :param z: redshift
        :type z: float >0
        :return: concentration parameter as float
        :rtype: float
        """
        return self.nfw_param.c_M_z(M, z)

    def gnfw_Mz(self, M, z, gamma_in):
        """Returns all needed parameter (in physical units modulo h) to draw the profile
        of the main halo r200 in physical Mpc/h rho_s in  h^2/Mpc^3 (physical) Rs in
        Mpc/h physical c unit less.

        :param M: Mass in physical M_sun/h
        :type M: float
        :param z: redshift
        :type z: float
        :param gamma_in: inner slope of the gNFW profile
        :type gamma_in: float
        :return: r200, rho0, c, Rs
        :rtype: float, float, float, float
        """
        c = self.c_M_z(M, z)
        r200 = self.r200_M(M, z)
        rho0 = self.rho0_c(c, z, gamma_in)
        Rs = r200 / c
        return r200, rho0, c, Rs



__author__ = "dangilman"

from lenstronomy.LensModel.lens_model import LensModel
import numpy as np


def setup_lens_model(lens_model, kwargs_lens, index_lens_split):
    """A method to split a lens model into a piece to vary or optimize, and fixed lens
    model whose deflection field will be interpolated. Note that this method currently
    only supports splitting the lens system at one particular redshift.

    :param lens_model: an instance of LensModel
    :param kwargs_lens: keyword arguments for the lens model
    :param index_lens_split: a list of indexes corresponding to the deflectors that will
        be left free to vary; every other deflector is assumed to remain static be
        absorbed into a net deflection field
    :return: an instance of LensModel corresponding to the fixed (static) deflectors, an
        instance of LensModel corresponding to free deflectors, keyword arguments for
        each lens model, source redshift, the redshift where the splitting occurs, and
        the background cosmology
    """
    z_source = lens_model.z_source
    cosmo = lens_model.cosmo
    lens_model_list = lens_model.lens_model_list
    lens_redshift_list = lens_model.redshift_list
    cosmo_bkg = lens_model.lens_model._multi_plane_base._cosmo_bkg
    lens_model_list_free = []
    lens_redshift_list_free = []
    kwargs_lens_free = []
    lens_model_list_fixed = []
    lens_redshift_list_fixed = []
    kwargs_lens_fixed = []
    for i in range(0, len(lens_model_list)):
        if i in index_lens_split:
            lens_model_list_free.append(lens_model_list[i])
            lens_redshift_list_free.append(lens_redshift_list[i])
            kwargs_lens_free.append(kwargs_lens[i])
        else:
            lens_model_list_fixed.append(lens_model_list[i])
            lens_redshift_list_fixed.append(lens_redshift_list[i])
            kwargs_lens_fixed.append(kwargs_lens[i])
    # for now we restrict to this relatively simple case
    for zi in lens_redshift_list_free:
        if zi != lens_redshift_list_free[0]:
            raise Exception("all free lens models must be at the same redshift")
    z_split = lens_redshift_list_free[0]
    lens_model_free = LensModel(
        lens_model_list_free,
        lens_redshift_list=lens_redshift_list_free,
        multi_plane=True,
        z_source=z_source,
        cosmo=cosmo,
    )
    lens_model_fixed = LensModel(
        lens_model_list_fixed,
        lens_redshift_list=lens_redshift_list_fixed,
        multi_plane=True,
        z_source=z_source,
        cosmo=cosmo,
    )
    return (
        lens_model_fixed,
        lens_model_free,
        kwargs_lens_fixed,
        kwargs_lens_free,
        z_source,
        z_split,
        cosmo_bkg,
    )


def setup_grids(
    grid_size, grid_resolution, coordinate_center_x=0.0, coordinate_center_y=0.0
):
    """Creates grids for use in the decoupled multiplane model.

    :param grid_size: The size (diameter of inscribed circle) of the grid
    :param grid_resolution: pixel scale (units arcsec / pixel)
    :param coordinate_center_x: center of the coordinate grid in arcsec
    :param coordinate_center_y: center of the coordinate grid in arcsec
    :return: 1d arrays of coordinates, tuple of 1d arrays of points defining the grid,
        number of pixels per axis
    """
    npix = int(grid_size / grid_resolution)
    if npix % 2 == 0:
        # we make sure this is odd so that grids include the center point
        npix += 1
    x = np.linspace(-grid_size / 2, grid_size / 2, npix)
    y = np.linspace(-grid_size / 2, grid_size / 2, npix)
    x += coordinate_center_x
    y += coordinate_center_y
    xx, yy = np.meshgrid(x, y)
    interp_points = (x, y)
    return xx.ravel(), yy.ravel(), interp_points, npix


def coordinates_and_deflections(
    lens_model_fixed,
    lens_model_free,
    kwargs_lens_fixed,
    kwargs_lens_free,
    x_coordinate_arcsec,
    y_coordinate_arcsec,
    z_split,
    z_source,
    cosmo_bkg,
):
    """Computes the lensed coordinates and deflection angles for the static lens model.

    :param lens_model_fixed: an instance of LensModel that is static
    :param lens_model_free: an instance of LensModel that is free to vary. NOTE: this
        should be a good guess of the "correct" lens model, as it will be used to
        estimate the coupling between the main deflector and deflectors between the main
        lens plane and the source plane
    :param kwargs_lens_fixed: keyword arguments for the fixed lens model
    :param kwargs_lens_free: keyword arguments for the free lens model
    :param x_coordinate_arcsec: coordinates on which to perform the interpolation;
        should be either a single coordinate, an array of coordiantes, or a list (or
        array) or coordinates corresponding to multiple images (see documentation in
        class_setup)
    :param y_coordinate_arcsec: coordinates on which to perform the interpolation;
        should be either a single coordinate, an array of coordiantes, or a list (or
        array) or coordinates corresponding to multiple images (see documentation in
        class_setup)
    :param z_split: the redshift where the free lens model lives
    :param z_source: the source redshift
    :param cosmo_bkg: background cosmology
    :return: comoving coordinates of light rays at z_split, foreground deflection
        angles, background deflections
    """

    Tds = cosmo_bkg.T_xy(z_split, z_source)
    Td = cosmo_bkg.T_xy(0, z_split)
    d_xy_source = cosmo_bkg.d_xy(0, z_source)
    d_xy_lens_source = cosmo_bkg.d_xy(z_split, z_source)
    reduced_to_phys = d_xy_source / d_xy_lens_source
    # first we handle all deflections up to the main lens plane, including substructure in the main lens plane
    (
        x_main_deflector,
        y_main_deflector,
        alpha_x_foreground,
        alpha_y_foreground,
    ) = lens_model_fixed.lens_model.ray_shooting_partial_comoving(
        np.zeros_like(x_coordinate_arcsec),
        np.zeros_like(y_coordinate_arcsec),
        x_coordinate_arcsec,
        y_coordinate_arcsec,
        0.0,
        z_split,
        kwargs_lens_fixed,
    )

    theta_x_main, theta_y_main = x_main_deflector / Td, y_main_deflector / Td
    alpha_x_main, alpha_y_main = lens_model_free.alpha(
        theta_x_main, theta_y_main, kwargs_lens_free
    )
    alpha_x_main *= reduced_to_phys
    alpha_y_main *= reduced_to_phys

    # get to the source plane
    angle_x = alpha_x_foreground - alpha_x_main
    angle_y = alpha_y_foreground - alpha_y_main
    x_source, y_source, _, _ = (
        lens_model_fixed.lens_model.ray_shooting_partial_comoving(
            x_main_deflector,
            y_main_deflector,
            angle_x,
            angle_y,
            z_split,
            z_source,
            kwargs_lens_fixed,
        )
    )

    # compute the effective deflection field for background halos
    alpha_x_background = (x_source - x_main_deflector) / Tds - angle_x
    alpha_y_background = (y_source - y_main_deflector) / Tds - angle_y
    return (
        x_main_deflector,
        y_main_deflector,
        alpha_x_foreground,
        alpha_y_foreground,
        alpha_x_background,
        alpha_y_background,
    )


def decoupled_multiplane_class_setup(
    lens_model_free,
    x,
    y,
    alpha_x_foreground,
    alpha_y_foreground,
    alpha_beta_subx,
    alpha_beta_suby,
    z_split,
    coordinate_type="POINT",
    interp_points=None,
    x_image=None,
    y_image=None,
    method="linear",
    bounds_error=False,
    fill_value=None,
):
    """This funciton creates the keyword arguments for a LensModel instance that is the
    decoupled multi-plane approxiamtion for the specified lens model.

    :param lens_model_free: the lens model with parameters free to vary
    :param x: comoving coordinate at z_split
    :param y: comoving coordinate at z_split
    :param alpha_x_foreground: ray angles at z_split (not including lens_model_free
        contribution)
    :param alpha_y_foreground: ray angles at z_split (not including lens_model_free
        contribution)
    :param alpha_beta_subx: deflection field from halos at redshift > z_split given the
        initial guess for the keyword arguments in lens_model_free
    :param alpha_beta_suby: deflection field from halos at redshift > z_split given the
        initial guess for the keyword arguments in lens_model_free
    :param z_split: redshift at which the lens model is decoupled from the line of sight
    :param coordinate_type: specifies the type of interpolation to use. Options are
        POINT, GRID, or MULTIPLE_IMAGES. POINT specifies a single point at which to
        compute the interpolation GRID specifies the interpolation on a regular grid
        MULTIPLE_IMAGES does interpolation on an array using the NEAREST method.
    :param lens_model_free:
    :param x: transverse comoving distance in x direction of the light rays at the main
        deflector
    :param y: transverse comoving distance in y direction of the light rays at the main
        deflector
    :param alpha_x_foreground: deflection angles from halos at redshift z<=z_split
    :param alpha_y_foreground: deflection angles from halos at redshift z<=z_split
    :param alpha_beta_subx: deflection angles from halos at redshift z > z_lens
    :param alpha_beta_suby: deflection angles from halos at redshift z > z_lens
    :param z_split: the redshift where foreground and background halos are split
    :param coordinate_type: a string specifying the type of coordinate of x. Options are
        GRID, POINT, and MULTIPLE_IMAGES
    :param interp_points: optional keyword argument passed to GRID method that specifies
        the interpolation grid
    :param x_image: optional keyword argument passed to multiple images argument that
        specifies the image coordinates
    :param y_image: optional keyword argument passed to multiple images argument that
        specifies the image coordinates
    :param method: the interpolation method used by RegularGridInterpolator if
        coordinate_type=='GRID'
    :param bounds_error: passed to RegularGridInterpolater, see documentation there
    :param fill_value: passed to RegularGridInterpolator, see documentation there
    :return: keyword arguments that can be passed into a LensModel class to create a
        decoupled-multiplane lens model
    """
    if coordinate_type == "GRID":
        from scipy.interpolate import RegularGridInterpolator

        npix = int(len(x) ** 0.5)
        interp_xD = RegularGridInterpolator(
            interp_points,
            x.reshape(npix, npix).T,
            bounds_error=bounds_error,
            fill_value=fill_value,
            method=method,
        )
        interp_yD = RegularGridInterpolator(
            interp_points,
            y.reshape(npix, npix).T,
            bounds_error=bounds_error,
            fill_value=fill_value,
            method=method,
        )
        interp_foreground_alpha_x = RegularGridInterpolator(
            interp_points,
            alpha_x_foreground.reshape(npix, npix).T,
            bounds_error=bounds_error,
            fill_value=fill_value,
            method=method,
        )
        interp_foreground_alpha_y = RegularGridInterpolator(
            interp_points,
            alpha_y_foreground.reshape(npix, npix).T,
            bounds_error=bounds_error,
            fill_value=fill_value,
            method=method,
        )
        interp_deltabeta_x = RegularGridInterpolator(
            interp_points,
            alpha_beta_subx.reshape(npix, npix).T,
            bounds_error=bounds_error,
            fill_value=fill_value,
            method=method,
        )
        interp_deltabeta_y = RegularGridInterpolator(
            interp_points,
            alpha_beta_suby.reshape(npix, npix).T,
            bounds_error=bounds_error,
            fill_value=fill_value,
            method=method,
        )
    elif coordinate_type == "POINT":
        interp_xD = lambda *args: x
        interp_yD = lambda *args: y
        interp_foreground_alpha_x = lambda *args: alpha_x_foreground
        interp_foreground_alpha_y = lambda *args: alpha_y_foreground
        interp_deltabeta_x = lambda *args: alpha_beta_subx
        interp_deltabeta_y = lambda *args: alpha_beta_suby

    elif coordinate_type == "MULTIPLE_IMAGES":
        from scipy.interpolate import NearestNDInterpolator

        interp_points = list(zip(x_image, y_image))
        interp_xD = NearestNDInterpolator(interp_points, x)
        interp_yD = NearestNDInterpolator(interp_points, y)
        interp_foreground_alpha_x = NearestNDInterpolator(
            interp_points, alpha_x_foreground
        )
        interp_foreground_alpha_y = NearestNDInterpolator(
            interp_points, alpha_y_foreground
        )
        interp_deltabeta_x = NearestNDInterpolator(interp_points, alpha_beta_subx)
        interp_deltabeta_y = NearestNDInterpolator(interp_points, alpha_beta_suby)

    else:
        raise Exception(
            "coordinate type must be either GRID, POINT, MULTIPLE_IMAGES, or MULTIPLE_IMAGES_GRID"
        )

    kwargs_decoupled_lens_model = {
        "x0_interp": interp_xD,
        "y0_interp": interp_yD,
        "alpha_x_interp_foreground": interp_foreground_alpha_x,
        "alpha_y_interp_foreground": interp_foreground_alpha_y,
        "alpha_x_interp_background": interp_deltabeta_x,
        "alpha_y_interp_background": interp_deltabeta_y,
        "z_split": z_split,
    }
    kwargs_lens_model = {
        "lens_model_list": lens_model_free.lens_model_list,
        "lens_redshift_list": lens_model_free.redshift_list,
        "cosmo": lens_model_free.cosmo,
        "multi_plane": True,
        "z_source": lens_model_free.z_source,
        "decouple_multi_plane": True,
        "kwargs_multiplane_model": kwargs_decoupled_lens_model,
    }
    return kwargs_lens_model


def setup_raytracing_lensmodels(
    x_image,
    y_image,
    lens_model,
    kwargs_lens,
    index_lens_split,
    grid_size,
    grid_resolution,
):
    """This function sets up the lens model used for high-resolution ray tracing with
    the decoupled multi-plane approximation.

    :param x_image: list of x-coordinates of lensed image
    :param y_image: list of y-coordinates of lensed image
    :param lens_model: an instance of LensModel
    :param kwargs_lens: keyword arguments for the lens model
    :param index_lens_split: list of integers specifying the lens models to be split
        from the line-of-sight population (see documentation in the DecoupledMultiplane
        class)
    :param grid_size: the size of the ray-tracing grid in arcsec
    :param grid_resolution: the resolution of the ray tracing grid in arcsec/pixel
    :return: a list of DecoupledMultiPlane lens models and corresponding keyword
        arguments
    """

    (
        lens_model_fixed,
        lens_model_free,
        kwargs_lens_fixed,
        kwargs_lens_free,
        z_source,
        z_split,
        cosmo_bkg,
    ) = setup_lens_model(lens_model, kwargs_lens, index_lens_split)
    kwargs_multiplane_lens_model_list = []
    multiplane_lens_model_list = []
    for image_index in range(0, len(x_image)):
        grid_x, grid_y, interp_points, npix = setup_grids(
            grid_size, grid_resolution, x_image[image_index], y_image[image_index]
        )
        (
            xD,
            yD,
            alpha_x_foreground,
            alpha_y_foreground,
            alpha_x_background,
            alpha_y_background,
        ) = coordinates_and_deflections(
            lens_model_fixed,
            lens_model_free,
            kwargs_lens_fixed,
            kwargs_lens_free,
            grid_x,
            grid_y,
            z_split,
            z_source,
            cosmo_bkg,
        )
        kwargs_multiplane_lens_model = decoupled_multiplane_class_setup(
            lens_model_free,
            xD,
            yD,
            alpha_x_foreground,
            alpha_y_foreground,
            alpha_x_background,
            alpha_y_background,
            z_split,
            coordinate_type="GRID",
            interp_points=interp_points,
        )
        kwargs_multiplane_lens_model_list.append(kwargs_multiplane_lens_model)
        multiplane_lens_model_list.append(LensModel(**kwargs_multiplane_lens_model))
    return multiplane_lens_model_list, kwargs_multiplane_lens_model_list



# -*- coding: utf-8 -*-
"""This module contains the class to compute lensing properties of a multi-Gaussian
convergence profile with the ellipticity defined in the convergence."""

__author__ = "ajshajib"

import numpy as np
from lenstronomy.LensModel.Profiles.gaussian_ellipse_kappa import GaussianEllipseKappa
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()

__all__ = ["MultiGaussianEllipseKappa"]


@export
class MultiGaussianEllipseKappa(LensProfileBase):
    """This class computes the lensing properties of a set of concentric elliptical
    Gaussian convergences."""

    param_names = ["amp", "sigma", "e1", "e2", "center_x", "center_y", "scale_factor"]
    lower_limit_default = {
        "amp": 0,
        "sigma": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
        "scale_factor": 0,
    }
    upper_limit_default = {
        "amp": 100,
        "sigma": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
        "scale_factor": 10000,
    }

    def __init__(self, use_scipy_wofz=True, min_ellipticity=1e-5):
        """

        :param use_scipy_wofz: To initiate ``class GaussianEllipseKappa``. If ``True``, Gaussian lensing will use ``scipy.special.wofz`` function. Set ``False`` for lower precision, but faster speed.
        :type use_scipy_wofz: ``bool``
        :param min_ellipticity: To be passed to ``class GaussianEllipseKappa``. Minimum ellipticity for Gaussian elliptical lensing calculation. For lower ellipticity than min_ellipticity the equations for the spherical case will be used.
        :type min_ellipticity: ``float``
        """
        self.gaussian_ellipse_kappa = GaussianEllipseKappa(
            use_scipy_wofz=use_scipy_wofz, min_ellipticity=min_ellipticity
        )
        super(MultiGaussianEllipseKappa, self).__init__()

    def function(
        self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0, scale_factor=1
    ):
        """Compute the potential function for a set of concentric elliptical Gaussian
        convergence profiles.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``numpy.array`` with ``dtype=float``
        :param sigma: Standard deviation of Gaussian
        :type sigma: ``numpy.array`` with ``dtype=float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordianate of centroid
        :type center_y: ``float``
        :param scale_factor: Scaling factor for amplitude
        :type scale_factor: ``float``
        :return: Potential for elliptical Gaussian convergence
        :rtype: ``float``, or ``numpy.array`` with ``shape = x.shape``
        """
        function = np.zeros_like(x, dtype=float)

        for i in range(len(amp)):
            if isinstance(e1, int) or isinstance(e1, float):
                e1_i = e1
                e2_i = e2
            else:
                e1_i = e1[i]
                e2_i = e2[i]
            function += self.gaussian_ellipse_kappa.function(
                x,
                y,
                scale_factor * amp[i],
                sigma[i],
                e1_i,
                e2_i,
                center_x,
                center_y,
            )
        return function

    def derivatives(
        self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0, scale_factor=1
    ):
        """Compute the derivatives of function angles :math:`\\partial f/\\partial x`,
        :math:`\\partial f/\\partial y` at :math:`x,\\ y` for a set of concentric
        elliptic Gaussian convergence profiles.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``numpy.array`` with ``dtype=float``
        :param sigma: Standard deviation of Gaussian
        :type sigma: ``numpy.array`` with ``dtype=float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordianate of centroid
        :type center_y: ``float``
        :param scale_factor: Scaling factor for amplitude
        :type scale_factor: ``float``
        :return: Deflection angle :math:`\\partial f/\\partial x`, :math:`\\partial f/\\partial y` for elliptical Gaussian convergence
        :rtype: tuple ``(float, float)`` or ``(numpy.array, numpy.array)`` with each ``numpy`` array's shape equal to ``x.shape``
        """
        f_x = np.zeros_like(x, dtype=float)
        f_y = np.zeros_like(x, dtype=float)

        for i in range(len(amp)):
            if isinstance(e1, int) or isinstance(e1, float):
                e1_i = e1
                e2_i = e2
            else:
                e1_i = e1[i]
                e2_i = e2[i]
            f_x_i, f_y_i = self.gaussian_ellipse_kappa.derivatives(
                x,
                y,
                amp=scale_factor * amp[i],
                sigma=sigma[i],
                e1=e1_i,
                e2=e2_i,
                center_x=center_x,
                center_y=center_y,
            )
            f_x += f_x_i
            f_y += f_y_i

        return f_x, f_y

    def hessian(self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0, scale_factor=1):
        """Compute Hessian matrix of function :math:`\\partial^2f/\\partial x^2`,
        :math:`\\partial^2 f/\\partial y^2`, :math:`\\partial^2 f/\\partial x\\partial
        y` for a set of concentric elliptic Gaussian convergence profiles.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``numpy.array`` with ``dtype=float``
        :param sigma: Standard deviation of Gaussian
        :type sigma: ``numpy.array`` with ``dtype=float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordianate of centroid
        :type center_y: ``float``
        :param scale_factor: Scaling factor for amplitude
         scale_factor: ``float``
        :return: Hessian :math:`\\partial^2f/\\partial x^2`, :math:`\\partial^2/\\partial x\\partial y`,
         :math:`\\partial^2/\\partial y\\partial x`, :math:`\\partial^2 f/\\partial y^2` for elliptical Gaussian convergence.
        :rtype: tuple ``(float, float, float)`` , or ``(numpy.array, numpy.array, numpy.array)``
         with each ``numpy`` array's shape equal to ``x.shape``
        """
        f_xx = np.zeros_like(x, dtype=float)
        f_yy = np.zeros_like(x, dtype=float)
        f_xy = np.zeros_like(x, dtype=float)

        for i in range(len(amp)):
            if isinstance(e1, int) or isinstance(e1, float):
                e1_i = e1
                e2_i = e2
            else:
                e1_i = e1[i]
                e2_i = e2[i]
            f_xx_i, f_xy_i, _, f_yy_i = self.gaussian_ellipse_kappa.hessian(
                x,
                y,
                amp=scale_factor * amp[i],
                sigma=sigma[i],
                e1=e1_i,
                e2=e2_i,
                center_x=center_x,
                center_y=center_y,
            )
            f_xx += f_xx_i
            f_yy += f_yy_i
            f_xy += f_xy_i

        return f_xx, f_xy, f_xy, f_yy

    def density_2d(
        self, x, y, amp, sigma, e1, e2, center_x=0, center_y=0, scale_factor=1
    ):
        """Compute the density of a set of concentric elliptical Gaussian convergence
        profiles :math:`\\sum A/(2\\pi \\sigma^2) \\exp(-( x^2+y^2/q^2)/2\\sigma^2)`.

        :param x: x coordinate
        :type x: ``float`` or ``numpy.array``
        :param y: y coordinate
        :type y: ``float`` or ``numpy.array``
        :param amp: Amplitude of Gaussian, convention: :math:`A/(2 \\pi\\sigma^2) \\exp(-(x^2+y^2/q^2)/2\\sigma^2)`
        :type amp: ``numpy.array`` with ``dtype=float``
        :param sigma: Standard deviation of Gaussian
        :type sigma: ``numpy.array`` with ``dtype=float``
        :param e1: Ellipticity parameter 1
        :type e1: ``float``
        :param e2: Ellipticity parameter 2
        :type e2: ``float``
        :param center_x: x coordinate of centroid
        :type center_x: ``float``
        :param center_y: y coordianate of centroid
        :type center_y: ``float``
        :param scale_factor: Scaling factor for amplitude
        :type scale_factor: ``float``
        :return: Density :math:`\\kappa` for elliptical Gaussian convergence
        :rtype: ``float``, or ``numpy.array`` with shape equal to ``x.shape``
        """
        density_2d = np.zeros_like(x, dtype=float)

        for i in range(len(amp)):
            if isinstance(e1, int) or isinstance(e1, float):
                e1_i = e1
                e2_i = e2
            else:
                e1_i = e1[i]
                e2_i = e2[i]
            density_2d += self.gaussian_ellipse_kappa.density_2d(
                x,
                y,
                amp=scale_factor * amp[i],
                sigma=sigma[i],
                e1=e1_i,
                e2=e2_i,
                center_x=center_x,
                center_y=center_y,
            )

        return density_2d



import numpy as np
import pytest
import numpy.testing as npt
from lenstronomy.Util import util
import lenstronomy.Util.param_util as param_util


def test_cart2polar():
    # singel 2d coordinate transformation
    center_x, center_y = 0, 0
    x = 1
    y = 1
    r, phi = param_util.cart2polar(x, y, center_x, center_y)
    assert r == np.sqrt(2)  # radial part
    assert phi == np.arctan(1)
    # array of 2d coordinates
    x = np.array([1, 2])
    y = np.array([1, 1])

    r, phi = param_util.cart2polar(x, y, center_x, center_y)
    assert r[0] == np.sqrt(2)  # radial part
    assert phi[0] == np.arctan(1)


def test_polar2cart():
    # singel 2d coordinate transformation
    center = np.array([0, 0])
    r = 1
    phi = np.pi
    x, y = param_util.polar2cart(r, phi, center)
    assert x == -1
    assert abs(y) < 10e-14


def test_phi_q2_ellipticity():
    phi, q = 0, 1
    e1, e2 = param_util.phi_q2_ellipticity(phi, q)
    assert e1 == 0
    assert e2 == 0

    phi, q = 1, 1
    e1, e2 = param_util.phi_q2_ellipticity(phi, q)
    assert e1 == 0
    assert e2 == 0

    phi, q = 2.0, 0.95
    e1, e2 = param_util.phi_q2_ellipticity(phi, q)
    npt.assert_almost_equal(e1, -0.016760092842656733, decimal=8)
    npt.assert_almost_equal(e2, -0.019405192187382792, decimal=8)

    phi, q = 0, 0.9
    e1, e2 = param_util.phi_q2_ellipticity(phi, q)
    npt.assert_almost_equal(e1, 0.05263157894736841, decimal=8)
    assert e2 == 0


def test_ellipticity2phi_q():
    e1, e2 = 0.3, 0
    phi, q = param_util.ellipticity2phi_q(e1, e2)
    assert phi == 0
    npt.assert_almost_equal(q, 0.53846153846153844, decimal=8)

    # Works on np arrays as well
    e1 = np.array([0.3, 0.9])
    e2 = np.array([0.0, 0.9])
    phi, q = param_util.ellipticity2phi_q(e1, e2)
    assert np.allclose(phi, [0.0, 0.39269908], atol=1.0e-08)
    assert np.allclose(q, [0.53846153, 5.00025001e-05], atol=1.0e-08)


def test_ellipticity2phi_q_symmetry():
    phi, q = 1.5, 0.8
    e1, e2 = param_util.phi_q2_ellipticity(phi, q)
    phi_new, q_new = param_util.ellipticity2phi_q(e1, e2)
    npt.assert_almost_equal(phi, phi_new, decimal=8)
    npt.assert_almost_equal(q, q_new, decimal=8)

    phi, q = -1.5, 0.8
    e1, e2 = param_util.phi_q2_ellipticity(phi, q)
    phi_new, q_new = param_util.ellipticity2phi_q(e1, e2)
    npt.assert_almost_equal(phi, phi_new, decimal=8)
    npt.assert_almost_equal(q, q_new, decimal=8)

    e1, e2 = 0.1, -0.1
    phi, q = param_util.ellipticity2phi_q(e1, e2)
    e1_new, e2_new = param_util.phi_q2_ellipticity(phi, q)
    npt.assert_almost_equal(e1, e1_new, decimal=10)
    npt.assert_almost_equal(e2, e2_new, decimal=10)

    e1, e2 = 2.99, -0.0
    phi, q = param_util.ellipticity2phi_q(e1, e2)
    e1_new, e2_new = param_util.phi_q2_ellipticity(phi, q)
    phi_new, q_new = param_util.ellipticity2phi_q(e1_new, e2_new)
    npt.assert_almost_equal(phi, phi_new, decimal=10)
    npt.assert_almost_equal(q, q_new, decimal=10)
    # npt.assert_almost_equal(e1, e1_new, decimal=10)
    # npt.assert_almost_equal(e2, e2_new, decimal=10)


def test_transform_e1e2():
    e1 = 0.01
    e2 = 0.0
    x = 0.0
    y = 1.0
    x_, y_ = param_util.transform_e1e2_product_average(
        x, y, e1, e2, center_x=0, center_y=0
    )
    x_new = (1 - e1) * x - e2 * y
    y_new = -e2 * x + (1 + e1) * y
    det = np.sqrt((1 - e1) * (1 + e1) + e2**2)
    npt.assert_almost_equal(x_, x_new / det, decimal=5)
    npt.assert_almost_equal(y_, y_new / det, decimal=5)


def test_transform_e1e2_product_average_new():
    x, y = util.make_grid(numPix=31, deltapix=1)

    e1, e2 = 0.3, 0.2
    center_x, center_y = 0, 0

    x_, y_ = param_util.transform_e1e2_product_average(x, y, e1, e2, center_x, center_y)
    r_new = np.sqrt(x_**2 + y_**2)

    x_old, y_old = param_util.transform_e1e2_product_average_old(
        x, y, e1, e2, center_x, center_y
    )
    r_old = np.sqrt(x_old**2 + y_old**2)
    npt.assert_almost_equal(r_new, r_old, decimal=8)


def test_phi_gamma_ellipticity():
    phi = -1.0
    gamma = 0.1
    e1, e2 = param_util.shear_polar2cartesian(phi, gamma)
    phi_out, gamma_out = param_util.shear_cartesian2polar(e1, e2)
    npt.assert_almost_equal(phi_out, phi, decimal=8)
    npt.assert_almost_equal(gamma_out, gamma_out, decimal=8)


def test_phi_gamma_ellipticity_2():
    e1, e2 = -0.04, -0.01
    phi, gamma = param_util.shear_cartesian2polar(e1, e2)

    e1_out, e2_out = param_util.shear_polar2cartesian(phi, gamma)
    npt.assert_almost_equal(e1, e1_out, decimal=10)
    npt.assert_almost_equal(e2, e2_out, decimal=10)


def test_displace_eccentricity():
    x, y = util.make_grid(numPix=10, deltapix=1)
    e1 = 0.1
    e2 = -0
    center_x, center_y = 0, 0
    x_, y_ = param_util.transform_e1e2_product_average(
        x, y, e1, e2, center_x=center_x, center_y=center_y
    )

    phi_G, q = param_util.ellipticity2phi_q(e1, e2)
    x_shift = x - center_x
    y_shift = y - center_y

    cos_phi = np.cos(phi_G)
    sin_phi = np.sin(phi_G)

    xt1 = cos_phi * x_shift + sin_phi * y_shift
    xt2 = -sin_phi * x_shift + cos_phi * y_shift
    xt1 *= np.sqrt(q)
    xt2 /= np.sqrt(q)
    npt.assert_almost_equal(x_, xt1, decimal=8)
    npt.assert_almost_equal(y_, xt2, decimal=8)

    x, y = np.array([1, 0]), np.array([0, 1])
    e1 = 0.1
    e2 = 0
    center_x, center_y = 0, 0
    x_, y_ = param_util.transform_e1e2_product_average(
        x, y, e1, e2, center_x=center_x, center_y=center_y
    )

    phi_G, q = param_util.ellipticity2phi_q(e1, e2)
    x_shift = x - center_x
    y_shift = y - center_y

    cos_phi = np.cos(phi_G)
    sin_phi = np.sin(phi_G)

    xt1 = cos_phi * x_shift + sin_phi * y_shift
    xt2 = -sin_phi * x_shift + cos_phi * y_shift
    xt1 *= np.sqrt(q)
    xt2 /= np.sqrt(q)
    npt.assert_almost_equal(x_, xt1, decimal=8)
    npt.assert_almost_equal(y_, xt2, decimal=8)


def test_transform_e1e2_square_average():
    x, y = np.array([1, 0]), np.array([0, 1])
    e1 = 0.1
    e2 = 0
    center_x, center_y = 0, 0

    x_, y_ = param_util.transform_e1e2_square_average(
        x, y, e1, e2, center_x=center_x, center_y=center_y
    )
    npt.assert_almost_equal(np.sum(x**2 + y**2), np.sum(x_**2 + y_**2), decimal=8)


def test_elliptical_distortion_product_average():
    """Tests that elliptical distortions in product average is the same as distortions
    under reduced shear.

    :return:
    """
    x, y = util.make_grid(numPix=20, deltapix=0.1)
    center_x, center_y = 1, -1

    (
        e1,
        e2,
    ) = (
        0.1,
        -0.2,
    )
    x_, y_ = param_util.elliptical_distortion_product_average(
        x, y, e1, e2, center_x, center_y
    )
    from lenstronomy.LensModel.lens_model import LensModel

    lens_model = LensModel(lens_model_list=["SHEAR_REDUCED"])
    kwargs_lens = [{"gamma1": e1, "gamma2": e2, "ra_0": center_x, "dec_0": center_y}]
    beta_x, beta_y = lens_model.ray_shooting(x, y, kwargs=kwargs_lens)
    npt.assert_almost_equal(x_, beta_x, decimal=5)
    npt.assert_almost_equal(y_, beta_y, decimal=5)


if __name__ == "__main__":
    pytest.main()



# Script that generates a a baseline JSON template following the COOLEST standard.
# The template is then used by the test routines in test/test_Util/test_COOLEST in order to cover the unit tests.
# Note the name of the file should *not* contain "test" to be ignored by pytest.


import os
import numpy as np
from astropy.io import fits
from coolest.template.lazy import *
from coolest.template.standard import COOLEST
from coolest.template.json import JSONSerializer

# from coolest.template.classes.probabilities import GaussianPrior
# from coolest.template.classes.parameter import PointEstimate
# from coolest.template.classes.probabilities import PosteriorStatistics


# basename of the generated COOLEST JSON template (no extension given here)
TEMPLATE_NAME = "coolest_template"

# purposedly introduce COOLEST options that are not yet supported by the lenstronomy's COOLEST interface
WITH_UNSUPPORTED_OPTION = True
UNSUPPORTED_OPTION_TYPE = "lenslight"  # "noise", "psf", "multiplane", "sourcelight", "lensmass", "lenslight", "massfield"


# Setup cosmology
cosmology = Cosmology(H0=73.0, Om0=0.3)

# Create a source galaxy represented by light profile
if WITH_UNSUPPORTED_OPTION and UNSUPPORTED_OPTION_TYPE == "sourcelight":
    main_source = Galaxy(
        "a source galaxy", 2.0, light_model=LightModel("PixelatedRegularGrid")
    )
else:
    main_source = Galaxy(
        "a source galaxy",
        2.0,
        light_model=LightModel("Sersic", "Shapelets", "LensedPS"),
    )

# Create a lens galaxy represented by a mass and light profiles
if WITH_UNSUPPORTED_OPTION and UNSUPPORTED_OPTION_TYPE == "lensmass":
    main_lens = Galaxy(
        "a lens galaxy",
        0.5,
        light_model=LightModel("Sersic"),
        mass_model=MassModel("PixelatedRegularGridPotential"),
    )
elif WITH_UNSUPPORTED_OPTION and UNSUPPORTED_OPTION_TYPE == "lenslight":
    main_lens = Galaxy(
        "a lens galaxy",
        0.5,
        light_model=LightModel("PixelatedRegularGrid"),
        mass_model=MassModel("SIE", "PEMD"),
    )
else:
    main_lens = Galaxy(
        "a lens galaxy",
        0.5,
        light_model=LightModel("Sersic"),
        mass_model=MassModel("SIE", "PEMD"),
    )

# Defines the external shear as a massive field
ext_shear = MassField(
    "an external shear",
    main_lens.redshift,
    mass_model=MassModel("ExternalShear", "ConvergenceSheet"),
)

# Put them in a list, which will also create unique IDs for each profile
if WITH_UNSUPPORTED_OPTION and UNSUPPORTED_OPTION_TYPE == "multiplane":
    lens_other_z = Galaxy(
        "a lens galaxy at different redshift",
        0.2,
        light_model=LightModel("Sersic"),
        mass_model=MassModel("SIE"),
    )
    source_other_z = Galaxy(
        "a source galaxy at different redshift", 1.3, light_model=LightModel("Sersic")
    )
    entity_list = LensingEntityList(
        ext_shear, main_lens, main_source, lens_other_z, source_other_z
    )
else:
    entity_list = LensingEntityList(ext_shear, main_lens, main_source)

# Define the origin of the coordinates system
origin = CoordinatesOrigin("00h11m20.244s", "-08d45m51.48s")

# EXAMPLE for accessing specific parameters and add priors/values/posteriors
# - add a gaussian prior to a given parameter
# main_lens.mass_model[0].parameters['gamma'].set_prior(GaussianPrior(mean=2.0, width=0.2))

# - add a point estimate to a given parameter
# ext_shear.mass_model[0].parameters['gamma_ext'].set_point_estimate(0.07)
# main_lens.light_model[1].parameters['q'].set_point_estimate(PointEstimate(value=0.89))

# - add a posterior distribution (as 0th and 1st order statistics)
# main_source.light_model[0].parameters['theta_eff'].set_posterior(PosteriorStatistics(mean=0.11, median=0.15, percentile_16th=0.03, percentile_84th=0.05))

# Define the data pixels and noise properties
data_shape = (100, 100)
obs_pixels = PixelatedRegularGrid(
    "obs.fits",
    field_of_view_x=[-4.0, 4.0],
    field_of_view_y=[-4.0, 4.0],
    num_pix_x=data_shape[0],
    num_pix_y=data_shape[1],
)
if WITH_UNSUPPORTED_OPTION and UNSUPPORTED_OPTION_TYPE == "noise":
    noise = UniformGaussianNoise(
        std_dev=0.005
    )  # other types of noise are supported in COOLEST
else:
    obs_noise_pixels = PixelatedRegularGrid(
        "noise_map.fits",
        field_of_view_x=[-4.0, 4.0],
        field_of_view_y=[-4.0, 4.0],
        num_pix_x=data_shape[0],
        num_pix_y=data_shape[1],
    )
    noise = NoiseMap(obs_noise_pixels)
observation = Observation(
    pixels=obs_pixels,
    noise=noise,  # other types of noise are supported in COOLEST
)

# Defines the instrument
if WITH_UNSUPPORTED_OPTION and UNSUPPORTED_OPTION_TYPE == "psf":
    psf = GaussianPSF(fwhm=0.8)
else:
    psf_shape = (23, 23)
    psf_pixels = PixelatedRegularGrid(
        "psf_kernel.fits",
        field_of_view_x=[-0.92, 0.92],
        field_of_view_y=[-0.92, 0.92],
        num_pix_x=psf_shape[0],
        num_pix_y=psf_shape[1],
    )
    psf = PixelatedPSF(psf_pixels)
instrument = Instrument(
    0.08,  # pixel size
    name="some instrument",
    readout_noise=4,
    band="F160W",
    psf=psf,  # other types of PSF are supported in COOLEST
)

# Master object for the standard
coolest = COOLEST("MAP", origin, entity_list, observation, instrument, cosmology)

# export as JSON file
output_filename = TEMPLATE_NAME
if WITH_UNSUPPORTED_OPTION:
    output_filename = "unsupp_" + output_filename + f"_{UNSUPPORTED_OPTION_TYPE}"
template_path = os.path.join(os.getcwd(), output_filename)
serializer = JSONSerializer(template_path, obj=coolest, check_external_files=True)
serializer.dump_simple()


# below we also create random fake FITS files that matches the above components
np.random.seed(0)
fits.writeto("obs.fits", np.random.randn(*data_shape), overwrite=True)
if not WITH_UNSUPPORTED_OPTION:
    fits.writeto("noise_map.fits", np.random.randn(*data_shape), overwrite=True)
    fits.writeto("psf_kernel.fits", np.random.randn(*psf_shape), overwrite=True)



__author__ = "sibirrer"

import numpy.testing as npt
import numpy as np
import pytest
import unittest

from lenstronomy.Analysis.kinematics_api import KinematicsAPI
import lenstronomy.Util.param_util as param_util
from astropy.cosmology import FlatLambdaCDM


class TestKinematicsAPI(object):
    def setup_method(self):
        pass

    def test_velocity_dispersion(self):
        z_lens = 0.5
        z_source = 1.5
        kwargs_model = {
            "lens_model_list": ["SPEP", "SHEAR", "SIS", "SIS", "SIS"],
            "lens_light_model_list": ["SERSIC_ELLIPSE", "SERSIC"],
        }

        theta_E = 1.5
        gamma = 1.8
        kwargs_lens = [
            {
                "theta_E": theta_E,
                "e1": 0,
                "center_x": -0.044798916793300093,
                "center_y": 0.0054408937891703788,
                "e2": 0,
                "gamma": gamma,
            },
            {"e1": -0.050871696555354479, "e2": -0.0061601733920590464},
            {
                "center_y": 2.79985456,
                "center_x": -2.32019894,
                "theta_E": 0.28165274714097904,
            },
            {
                "center_y": 3.83985426,
                "center_x": -2.32019933,
                "theta_E": 0.0038110812674654873,
            },
            {
                "center_y": 4.31985428,
                "center_x": -1.68019931,
                "theta_E": 0.45552039839735037,
            },
        ]

        phi, q = -0.52624727893702705, 0.79703498156919605
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_lens_light = [
            {
                "n_sersic": 1.1212528655709217,
                "center_x": -0.019674496231393473,
                "e1": e1,
                "e2": e2,
                "amp": 1.1091367792010356,
                "center_y": 0.076914975081560991,
                "R_sersic": 0.42691611878867058,
            },
            {
                "R_sersic": 0.03025682660635394,
                "amp": 139.96763298885992,
                "n_sersic": 1.90000008624093865,
                "center_x": -0.019674496231393473,
                "center_y": 0.076914975081560991,
            },
        ]
        r_ani = 0.62
        kwargs_anisotropy = {"r_ani": r_ani}
        R_slit = 3.8
        dR_slit = 1.0
        aperture_type = "slit"
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "center_ra": 0,
            "width": dR_slit,
            "length": R_slit,
            "angle": 0,
            "center_dec": 0,
        }

        psf_fwhm = 0.7
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        anisotropy_model = "OM"
        kwargs_mge = {"n_comp": 20}
        r_eff = 0.211919902322
        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            lens_model_kinematics_bool=[True, False, False, False, False],
            anisotropy_model=anisotropy_model,
            kwargs_mge_light=kwargs_mge,
            kwargs_mge_mass=kwargs_mge,
            sampling_number=1000,
            MGE_light=True,
        )

        v_sigma = kinematicAPI.velocity_dispersion(
            kwargs_lens, kwargs_lens_light, kwargs_anisotropy, r_eff=r_eff
        )

        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            lens_model_kinematics_bool=[True, False, False, False, False],
            anisotropy_model=anisotropy_model,
            kwargs_mge_light=kwargs_mge,
            kwargs_mge_mass=kwargs_mge,
            sampling_number=1000,
            MGE_light=True,
            MGE_mass=True,
        )
        v_sigma_mge_lens = kinematicAPI.velocity_dispersion(
            kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
        )
        # v_sigma_mge_lens = kinematicAPI.velocity_dispersion_numerical(kwargs_lens, kwargs_lens_light, kwargs_anisotropy, kwargs_aperture,
        #                                                          kwargs_psf, anisotropy_model, MGE_light=True, MGE_mass=True, theta_E=theta_E,
        #                                                          kwargs_mge_light=kwargs_mge, kwargs_mge_mass=kwargs_mge,
        #                                                          r_eff=r_eff)
        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            lens_model_kinematics_bool=[True, False, False, False, False],
            anisotropy_model=anisotropy_model,
            kwargs_mge_light=kwargs_mge,
            kwargs_mge_mass=kwargs_mge,
            sampling_number=1000,
            MGE_light=False,
            MGE_mass=False,
            Hernquist_approx=True,
        )
        v_sigma_hernquist = kinematicAPI.velocity_dispersion(
            kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
        )
        # v_sigma_hernquist = kinematicAPI.velocity_dispersion_numerical(kwargs_lens, kwargs_lens_light, kwargs_anisotropy,
        #                                                          kwargs_aperture, kwargs_psf, anisotropy_model,
        #                                                          MGE_light=False, MGE_mass=False,
        #                                                          r_eff=r_eff, Hernquist_approx=True)

        vel_disp_temp = kinematicAPI.velocity_dispersion_analytical(
            theta_E, gamma, r_ani=r_ani, r_eff=r_eff
        )
        print(v_sigma, vel_disp_temp)
        # assert 1 == 0
        npt.assert_almost_equal(v_sigma / vel_disp_temp, 1, decimal=1)
        npt.assert_almost_equal(v_sigma_mge_lens / v_sigma, 1, decimal=1)
        npt.assert_almost_equal(v_sigma / v_sigma_hernquist, 1, decimal=1)

    def test_galkin_settings(self):
        z_lens = 0.5
        z_source = 1.5
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "lens_light_model_list": ["HERNQUIST"],
        }

        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        kwargs_lens_light = [{"amp": 1, "Rs": 1, "center_x": 0, "center_y": 0}]
        r_ani = 0.62
        kwargs_anisotropy = {"r_ani": r_ani}
        R_slit = 3.8
        dR_slit = 1.0
        aperture_type = "slit"
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "center_ra": 0,
            "width": dR_slit,
            "length": R_slit,
            "angle": 0,
            "center_dec": 0,
        }

        psf_fwhm = 0.7
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        anisotropy_model = "OM"
        kwargs_mge = {"n_comp": 20}
        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            analytic_kinematics=True,
            anisotropy_model=anisotropy_model,
            kwargs_mge_light=kwargs_mge,
            kwargs_mge_mass=kwargs_mge,
            sampling_number=1000,
        )
        galkin, kwargs_profile, kwargs_light = kinematicAPI.galkin_settings(
            kwargs_lens, kwargs_lens_light, r_eff=None, theta_E=None, gamma=None
        )
        npt.assert_almost_equal(kwargs_profile["gamma"], 2, decimal=2)

        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture=[kwargs_aperture],
            kwargs_seeing=[kwargs_psf],
            analytic_kinematics=True,
            anisotropy_model=anisotropy_model,
            multi_observations=True,
            kwargs_mge_light=kwargs_mge,
            kwargs_mge_mass=kwargs_mge,
            sampling_number=1000,
        )
        galkin, kwargs_profile, kwargs_light = kinematicAPI.galkin_settings(
            kwargs_lens, kwargs_lens_light, r_eff=None, theta_E=None, gamma=None
        )
        npt.assert_almost_equal(kwargs_profile["gamma"], 2, decimal=2)

    def test_kinematic_light_profile(self):
        z_lens = 0.5
        z_source = 1.5
        kwargs_options = {"lens_light_model_list": ["HERNQUIST_ELLIPSE", "SERSIC"]}
        kwargs_mge = {"n_comp": 20}
        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_options,
            kwargs_seeing={},
            kwargs_aperture={},
            anisotropy_model="OM",
        )
        r_eff = 0.2
        kwargs_lens_light = [
            {
                "amp": 1,
                "Rs": r_eff * 0.551,
                "e1": 0.0,
                "e2": 0,
                "center_x": 0,
                "center_y": 0,
            },
            {"amp": 1, "R_sersic": 1, "n_sersic": 2, "center_x": -10, "center_y": -10},
        ]
        light_profile_list, kwargs_light = kinematicAPI.kinematic_light_profile(
            kwargs_lens_light,
            MGE_fit=True,
            r_eff=r_eff,
            model_kinematics_bool=[True, False],
            kwargs_mge=kwargs_mge,
        )
        assert light_profile_list[0] == "MULTI_GAUSSIAN"

        light_profile_list, kwargs_light = kinematicAPI.kinematic_light_profile(
            kwargs_lens_light,
            MGE_fit=False,
            r_eff=r_eff,
            model_kinematics_bool=[True, False],
        )
        assert light_profile_list[0] == "HERNQUIST_ELLIPSE"

        light_profile_list, kwargs_light = kinematicAPI.kinematic_light_profile(
            kwargs_lens_light,
            MGE_fit=False,
            Hernquist_approx=True,
            r_eff=r_eff,
            model_kinematics_bool=[True, False],
        )
        assert light_profile_list[0] == "HERNQUIST"
        npt.assert_almost_equal(
            kwargs_light[0]["Rs"] / kwargs_lens_light[0]["Rs"], 1, decimal=2
        )

    def test_kinematic_lens_profiles(self):
        z_lens = 0.5
        z_source = 1.5
        kwargs_options = {"lens_model_list": ["SPEP", "SHEAR"]}
        kin_api = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_options,
            kwargs_aperture={},
            kwargs_seeing={},
            anisotropy_model="OM",
        )
        kwargs_lens = [
            {
                "theta_E": 1.4272358196260446,
                "e1": 0,
                "center_x": -0.044798916793300093,
                "center_y": 0.0054408937891703788,
                "e2": 0,
                "gamma": 1.8,
            },
            {"e1": -0.050871696555354479, "e2": -0.0061601733920590464},
        ]

        kwargs_mge = {"n_comp": 20}
        mass_profile_list, kwargs_profile = kin_api.kinematic_lens_profiles(
            kwargs_lens,
            MGE_fit=True,
            kwargs_mge=kwargs_mge,
            theta_E=1.4,
            model_kinematics_bool=[True, False],
        )
        assert mass_profile_list[0] == "MULTI_GAUSSIAN"

        mass_profile_list, kwargs_profile = kin_api.kinematic_lens_profiles(
            kwargs_lens, MGE_fit=False, model_kinematics_bool=[True, False]
        )
        assert mass_profile_list[0] == "SPEP"

    def test_model_dispersion(self):
        np.random.seed(42)
        z_lens = 0.5
        z_source = 1.5
        r_eff = 1.0
        theta_E = 1.0
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "lens_light_model_list": ["HERNQUIST"],
        }
        kwargs_lens = [{"theta_E": theta_E, "center_x": 0, "center_y": 0}]
        kwargs_lens_light = [
            {"amp": 1, "Rs": r_eff * 0.551, "center_x": 0, "center_y": 0}
        ]
        kwargs_anisotropy = {"r_ani": 1}
        # settings

        R_slit = 3.8
        dR_slit = 1.0
        aperture_type = "slit"
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "center_ra": 0,
            "width": dR_slit,
            "length": R_slit,
            "angle": 0,
            "center_dec": 0,
        }
        psf_fwhm = 0.7
        kwargs_seeing = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        anisotropy_model = "OM"
        kin_api = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture,
            kwargs_seeing,
            anisotropy_model=anisotropy_model,
        )

        kwargs_numerics_galkin = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 1000,
            "min_integrate": 0.0001,
        }
        kin_api.kinematics_modeling_settings(
            anisotropy_model,
            kwargs_numerics_galkin,
            analytic_kinematics=True,
            Hernquist_approx=False,
            MGE_light=False,
            MGE_mass=False,
        )
        vel_disp_analytic = kin_api.velocity_dispersion(
            kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
            gamma=2,
        )

        kin_api.kinematics_modeling_settings(
            anisotropy_model,
            kwargs_numerics_galkin,
            analytic_kinematics=False,
            Hernquist_approx=False,
            MGE_light=False,
            MGE_mass=False,
        )
        vel_disp_numerical = kin_api.velocity_dispersion(
            kwargs_lens, kwargs_lens_light, kwargs_anisotropy
        )  # ,
        # r_eff=r_eff, theta_E=theta_E, gamma=2)
        npt.assert_almost_equal(vel_disp_numerical / vel_disp_analytic, 1, decimal=2)

        kin_api.kinematics_modeling_settings(
            anisotropy_model,
            kwargs_numerics_galkin,
            analytic_kinematics=False,
            Hernquist_approx=False,
            MGE_light=False,
            MGE_mass=False,
            kwargs_mge_light={"n_comp": 10},
            kwargs_mge_mass={"n_comp": 5},
        )
        assert kin_api._kwargs_mge_mass["n_comp"] == 5
        assert kin_api._kwargs_mge_light["n_comp"] == 10

    def test_velocity_dispersion_map_direct_convolved_against_jampy(self):
        """Test the computed velocity dispersion map through the Kinematics_API with PSF
        convolution against `jampy` computed values.

        The `jampy` values are computed using the same model, grid, and PSF used for
        Galkin using the code below:

        .. code-block:: python

            import numpy as np
            from astropy.cosmology import FlatLambdaCDM
            from lenstronomy.LightModel.light_model import LightModel
            from lenstronomy.LensModel.lens_model import LensModel
            from jampy.jam_axi_proj import jam_axi_proj
            from mgefit.mge_fit_1d import mge_fit_1d

            z_l = 0.3
            z_s = 0.7

            pixel_size = 0.1457
            x_grid, y_grid = np.meshgrid(
                np.arange(-3.0597, 3.1597, pixel_size),
                np.arange(-3.0597, 3.1597, pixel_size),
            )
            psf_fwhm = 0.7

            light_model = LightModel(["SERSIC"])
            kwargs_lens_light = [
                {
                    "amp": 0.09,
                    "R_sersic": 1.2,
                    "n_sersic": 0.9,
                    "center_x": 0.0,
                    "center_y": 0.0,
                }
            ]

            rs = np.logspace(-2.5, 2, 300)
            flux_r = light_model.surface_brightness(rs, 0 * rs, kwargs_lens_light)

            mge_fit = mge_fit_1d(rs, flux_r, ngauss=20, quiet=True)
            sigma_lum = mge_fit.sol[1]
            surf_lum = mge_fit.sol[0] / (np.sqrt(2 * np.pi) * sigma_lum)
            qobs_lum = np.ones_like(sigma_lum)


            lens_model = LensModel(["EPL"])
            kwargs_lens = [
                {
                    "theta_E": 1.63,
                    "gamma": 2.02,
                    "e1": 0.0,
                    "e2": 0.0,
                    "center_x": 0.0,
                    "center_y": 0.0,
                }
            ]

            mass_r = lens_model.kappa(rs, rs * 0, kwargs_lens)
            mass_mge = mge_fit_1d(rs, mass_r, ngauss=20, quiet=True, plot=False)

            cosmo = FlatLambdaCDM(H0=70, Om0=0.3, Ob0=0.05)
            D_d = cosmo.angular_diameter_distance(z_l).value
            D_s = cosmo.angular_diameter_distance(z_s).value
            D_ds = cosmo.angular_diameter_distance_z1z2(z_l, z_s).value
            c2_4piG = 1.6624541593797972e6
            sigma_crit = c2_4piG * D_s / D_ds / D_d

            sigma_pot = mass_mge.sol[1]
            surf_pot = mass_mge.sol[0] / (np.sqrt(2 * np.pi) * sigma_pot) * sigma_crit
            qobs_pot = np.ones_like(sigma_pot)

            bs = np.ones_like(surf_lum) * 0.25

            jam = jam_axi_proj(
                surf_lum,
                sigma_lum,
                qobs_lum,
                surf_pot,
                sigma_pot,
                qobs_pot,
                inc=90,
                mbh=0,
                distance=D_d,
                xbin=x_grid.flatten(),
                ybin=y_grid.flatten(),
                plot=False,
                pixsize=pixel_size,
                pixang=0,
                quiet=1,
                sigmapsf=psf_fwhm / 2.355,
                normpsf=1,
                moment="zz",
                align="sph",
                beta=bs,
                ml=1,
            ).model

            jampy_vel_dis = jam.reshape(x_grid.shape)[14:28, 14:28]
        """
        z_l = 0.3
        z_s = 0.7

        anisotropy_type = "const"

        kwargs_model = {
            "lens_model_list": ["EPL"],
            "lens_light_model_list": ["SERSIC"],
        }

        pixel_size = 0.1457
        x_grid, y_grid = np.meshgrid(
            np.arange(-3.0597, 3.1597, pixel_size),  # x-axis points to negative RA
            np.arange(-3.0597, 3.1597, pixel_size),
        )
        psf_fwhm = 0.7

        kwargs_aperture = {
            "aperture_type": "IFU_grid",
            "x_grid": x_grid,
            "y_grid": y_grid,
        }
        kwargs_seeing = {
            "psf_type": "GAUSSIAN",
            "fwhm": psf_fwhm,
        }

        kwargs_galkin_numerics = {
            "interpol_grid_num": 1000,
            "log_integration": True,
            "max_integrate": 100,
            "min_integrate": 0.001,
        }

        light_model_bool = [True]
        lens_model_bool = [True]

        cosmo = FlatLambdaCDM(H0=70, Om0=0.3, Ob0=0.05)

        kinematics_api = KinematicsAPI(
            z_lens=z_l,
            z_source=z_s,
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_seeing,
            anisotropy_model=anisotropy_type,
            cosmo=cosmo,
            lens_model_kinematics_bool=lens_model_bool,
            light_model_kinematics_bool=light_model_bool,
            multi_observations=False,
            kwargs_numerics_galkin=kwargs_galkin_numerics,
            analytic_kinematics=False,
            Hernquist_approx=False,
            MGE_light=True,
            MGE_mass=False,
            kwargs_mge_light=None,
            kwargs_mge_mass=None,
            sampling_number=1000,
            num_kin_sampling=2000,
            num_psf_sampling=500,
        )

        beta = 0.25

        kwargs_lens = [
            {
                "theta_E": 1.63,
                "gamma": 2.02,
                "e1": 0.0,
                "e2": 0.0,
                "center_x": 0.0,
                "center_y": 0.0,
            }
        ]

        kwargs_lens_light = [
            {
                "amp": 0.09,
                "R_sersic": 1.2,
                "n_sersic": 0.9,
                "center_x": 0.0,
                "center_y": 0.0,
            },
        ]

        kwargs_anisotropy = {"beta": beta}

        vel_dis = kinematics_api.velocity_dispersion_map(
            kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=kwargs_lens_light[0]["R_sersic"],
            theta_E=kwargs_lens[0]["theta_E"],
            gamma=kwargs_lens[0]["gamma"],
            kappa_ext=0,
            direct_convolve=True,
            supersampling_factor=5,
            voronoi_bins=None,
        )

        jampy_vel_dis = np.array(
            [
                [
                    244.34442569258272,
                    251.77191207554029,
                    259.1237317425143,
                    266.0992101486318,
                    272.2922431213921,
                    277.22364790181547,
                    280.420765688001,
                    281.5303313645199,
                    280.4207656880009,
                    277.2236479018153,
                    272.29224312139183,
                    266.0992101486315,
                    259.12373174251394,
                    251.77191207553997,
                ],
                [
                    251.80156000934159,
                    260.6357397539551,
                    269.61178853012774,
                    278.3628625494463,
                    286.33644100586974,
                    292.8270286488553,
                    297.10211203395255,
                    298.59802357565803,
                    297.10211203395244,
                    292.82702864885493,
                    286.33644100586935,
                    278.3628625494459,
                    269.61178853012734,
                    260.63573975395474,
                ],
                [
                    259.18185345539825,
                    269.640505763082,
                    280.56771429629146,
                    291.5309223361145,
                    301.78703955388204,
                    310.3144827300153,
                    316.012479656365,
                    318.0205681790704,
                    316.0124796563648,
                    310.3144827300149,
                    301.7870395538816,
                    291.53092233611403,
                    280.5677142962909,
                    269.64050576308153,
                ],
                [
                    266.1833842428864,
                    278.41798325669726,
                    291.5574149415321,
                    305.1039738960561,
                    318.07633459266276,
                    329.0488408271573,
                    336.458572352135,
                    339.08263292163014,
                    336.45857235213475,
                    329.04884082715677,
                    318.07633459266214,
                    305.1039738960555,
                    291.5574149415314,
                    278.41798325669674,
                ],
                [
                    272.39862829693527,
                    286.41407229509963,
                    301.83604121785453,
                    318.0987201357714,
                    333.9475055375945,
                    347.5046489029379,
                    356.71364962756127,
                    359.9823490359285,
                    356.71364962756104,
                    347.50464890293733,
                    333.9475055375937,
                    318.09872013577063,
                    301.83604121785373,
                    286.41407229509906,
                ],
                [
                    277.3470251596913,
                    292.9218044651196,
                    310.3804617235422,
                    329.08788077556756,
                    347.5210760854689,
                    363.38044693017576,
                    374.17592141954464,
                    378.00928078574117,
                    374.1759214195443,
                    363.3804469301751,
                    347.521076085468,
                    329.0878807755667,
                    310.3804617235413,
                    292.921804465119,
                ],
                [
                    280.5548042286683,
                    297.2075758118116,
                    316.08891485916615,
                    336.5076978971311,
                    356.73985052453725,
                    374.1855594337415,
                    386.0630046075378,
                    390.27902142174673,
                    386.06300460753744,
                    374.1855594337407,
                    356.7398505245362,
                    336.50769789713013,
                    316.08891485916513,
                    297.2075758118109,
                ],
                [
                    281.6680018133935,
                    298.7071114771313,
                    318.10051915586695,
                    339.13510983521377,
                    360.01175680123686,
                    378.0220486491648,
                    390.28213180326384,
                    394.6329544633877,
                    390.28213180326344,
                    378.022048649164,
                    360.01175680123583,
                    339.13510983521275,
                    318.1005191558659,
                    298.70711147713064,
                ],
                [
                    280.55480422866816,
                    297.20757581181147,
                    316.088914859166,
                    336.5076978971309,
                    356.73985052453696,
                    374.1855594337412,
                    386.06300460753744,
                    390.27902142174634,
                    386.06300460753704,
                    374.1855594337404,
                    356.73985052453594,
                    336.5076978971299,
                    316.08891485916496,
                    297.2075758118107,
                ],
                [
                    277.3470251596911,
                    292.92180446511935,
                    310.3804617235418,
                    329.0878807755671,
                    347.52107608546834,
                    363.3804469301751,
                    374.17592141954384,
                    378.0092807857404,
                    374.17592141954356,
                    363.38044693017434,
                    347.5210760854674,
                    329.0878807755662,
                    310.3804617235409,
                    292.92180446511867,
                ],
                [
                    272.398628296935,
                    286.4140722950993,
                    301.8360412178541,
                    318.0987201357708,
                    333.94750553759377,
                    347.50464890293705,
                    356.7136496275603,
                    359.9823490359275,
                    356.71364962756,
                    347.5046489029364,
                    333.947505537593,
                    318.09872013577,
                    301.8360412178533,
                    286.4140722950987,
                ],
                [
                    266.1833842428861,
                    278.4179832566968,
                    291.55741494153153,
                    305.10397389605544,
                    318.07633459266196,
                    329.0488408271564,
                    336.45857235213396,
                    339.08263292162917,
                    336.45857235213373,
                    329.0488408271559,
                    318.07633459266134,
                    305.10397389605487,
                    291.55741494153085,
                    278.41798325669635,
                ],
                [
                    259.1818534553979,
                    269.64050576308153,
                    280.5677142962909,
                    291.53092233611386,
                    301.78703955388124,
                    310.31448273001433,
                    316.01247965636395,
                    318.02056817906936,
                    316.01247965636384,
                    310.314482730014,
                    301.78703955388073,
                    291.5309223361133,
                    280.56771429629026,
                    269.6405057630811,
                ],
                [
                    251.80156000934133,
                    260.6357397539548,
                    269.61178853012734,
                    278.36286254944577,
                    286.33644100586923,
                    292.82702864885465,
                    297.1021120339519,
                    298.59802357565735,
                    297.10211203395176,
                    292.82702864885437,
                    286.33644100586883,
                    278.36286254944537,
                    269.61178853012683,
                    260.6357397539544,
                ],
            ]
        )

        assert np.max(np.abs(jampy_vel_dis / vel_dis[14:28, 14:28] - 1)) < 0.009

    def test_velocity_dispersion_map(self):
        np.random.seed(42)
        z_lens = 0.5
        z_source = 1.5
        kwargs_options = {
            "lens_model_list": ["SIS"],
            "lens_light_model_list": ["HERNQUIST"],
        }
        r_eff = 1.0
        theta_E = 1
        kwargs_lens = [{"theta_E": theta_E, "center_x": 0, "center_y": 0}]
        kwargs_lens_light = [
            {"amp": 1, "Rs": r_eff * 0.551, "center_x": 0, "center_y": 0}
        ]
        kwargs_anisotropy = {"r_ani": 1}

        r_bins = np.array([0, 0.5, 1])
        aperture_type = "IFU_shells"
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "center_ra": 0,
            "r_bins": r_bins,
            "center_dec": 0,
        }
        psf_fwhm = 0.7
        kwargs_seeing = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        anisotropy_model = "OM"
        kin_api = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_options,
            kwargs_aperture=kwargs_aperture,
            kwargs_seeing=kwargs_seeing,
            anisotropy_model=anisotropy_model,
        )

        kwargs_numerics_galkin = {
            "interpol_grid_num": 500,
            "log_integration": True,
            "max_integrate": 10,
            "min_integrate": 0.001,
        }
        kin_api.kinematics_modeling_settings(
            anisotropy_model,
            kwargs_numerics_galkin,
            analytic_kinematics=True,
            Hernquist_approx=False,
            MGE_light=False,
            MGE_mass=False,
            num_kin_sampling=1000,
            num_psf_sampling=100,
        )
        vel_disp_analytic = kin_api.velocity_dispersion_map(
            kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
            gamma=2,
        )

        kin_api.kinematics_modeling_settings(
            anisotropy_model,
            kwargs_numerics_galkin,
            analytic_kinematics=False,
            Hernquist_approx=False,
            MGE_light=False,
            MGE_mass=False,
            num_kin_sampling=1000,
            num_psf_sampling=100,
        )
        vel_disp_numerical = kin_api.velocity_dispersion_map(
            kwargs_lens,
            kwargs_lens_light,
            kwargs_anisotropy,
            r_eff=r_eff,
            theta_E=theta_E,
            gamma=2,
        )
        print(vel_disp_numerical, vel_disp_analytic)
        npt.assert_almost_equal(vel_disp_numerical, vel_disp_analytic, decimal=-1)

        z_lens = 0.5
        z_source = 1.5
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "lens_light_model_list": ["HERNQUIST"],
        }

        xs, ys = np.meshgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20))
        kwargs_aperture = {
            "aperture_type": "IFU_grid",
            "x_grid": xs,
            "y_grid": ys,
        }
        kwargs_seeing = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        kin_api = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_seeing=kwargs_seeing,
            kwargs_aperture=kwargs_aperture,
            anisotropy_model="OM",
        )
        kin_api.velocity_dispersion_map(
            [{"theta_E": 1, "center_x": 0, "center_y": 0}],
            [{"Rs": 1, "amp": 1, "center_x": 0, "center_y": 0}],
            {"r_ani": 1},
            direct_convolve=False,
        )

        kin_api.velocity_dispersion_map(
            [{"theta_E": 1, "center_x": 0, "center_y": 0}],
            [{"Rs": 1, "amp": 1, "center_x": 0, "center_y": 0}],
            {"r_ani": 1},
            direct_convolve=True,
        )

    def test_interpolated_sersic(self):
        from lenstronomy.Analysis.light2mass import light2mass_interpol

        kwargs_light = [
            {
                "n_sersic": 2,
                "R_sersic": 0.5,
                "amp": 1,
                "center_x": 0.01,
                "center_y": 0.01,
            }
        ]
        kwargs_lens = [
            {
                "n_sersic": 2,
                "R_sersic": 0.5,
                "k_eff": 1,
                "center_x": 0.01,
                "center_y": 0.01,
            }
        ]
        deltaPix = 0.1
        numPix = 100

        kwargs_interp = light2mass_interpol(
            ["SERSIC"],
            kwargs_lens_light=kwargs_light,
            numPix=numPix,
            deltaPix=deltaPix,
            subgrid_res=5,
        )
        kwargs_lens_interp = [kwargs_interp]
        from lenstronomy.Analysis.kinematics_api import KinematicsAPI

        z_lens = 0.5
        z_source = 1.5
        r_ani = 0.62
        kwargs_anisotropy = {"r_ani": r_ani}
        R_slit = 3.8
        dR_slit = 1.0
        aperture_type = "slit"
        kwargs_aperture = {
            "center_ra": 0,
            "width": dR_slit,
            "length": R_slit,
            "angle": 0,
            "center_dec": 0,
            "aperture_type": aperture_type,
        }
        psf_fwhm = 0.7
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        anisotropy_model = "OM"
        r_eff = 0.5
        kwargs_model = {
            "lens_model_list": ["SERSIC"],
            "lens_light_model_list": ["SERSIC"],
        }
        kwargs_mge = {"n_comp": 20}
        kinematic_api = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            anisotropy_model=anisotropy_model,
            MGE_light=True,
            MGE_mass=True,
            kwargs_mge_mass=kwargs_mge,
            kwargs_mge_light=kwargs_mge,
        )

        v_sigma = kinematic_api.velocity_dispersion(
            kwargs_lens, kwargs_light, kwargs_anisotropy, r_eff=r_eff, theta_E=1
        )
        kwargs_model_interp = {
            "lens_model_list": ["INTERPOL"],
            "lens_light_model_list": ["SERSIC"],
        }
        kinematic_api_interp = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model_interp,
            kwargs_aperture,
            kwargs_seeing=kwargs_psf,
            anisotropy_model=anisotropy_model,
            MGE_light=True,
            MGE_mass=True,
            kwargs_mge_mass=kwargs_mge,
            kwargs_mge_light=kwargs_mge,
        )
        v_sigma_interp = kinematic_api_interp.velocity_dispersion(
            kwargs_lens_interp,
            kwargs_light,
            kwargs_anisotropy,
            theta_E=1.0,
            r_eff=r_eff,
        )
        npt.assert_almost_equal(v_sigma / v_sigma_interp, 1, 1)
        # use as kinematic constraints
        # compare with MGE Sersic kinematic estimate

    def test_copy_centers(self):
        z_lens = 0.5
        z_source = 1.5
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "lens_light_model_list": ["HERNQUIST"],
        }
        kwargs_aperture = {
            "aperture_type": "slit",
            "center_ra": 0,
            "width": 1,
            "length": 1,
            "angle": 0,
            "center_dec": 0,
        }
        psf_fwhm = 0.7
        kwargs_seeing = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        anisotropy_model = "OM"
        kinematicAPI = KinematicsAPI(
            z_lens,
            z_source,
            kwargs_model,
            kwargs_seeing,
            kwargs_aperture,
            anisotropy_model=anisotropy_model,
            analytic_kinematics=True,
        )
        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        kwargs_lens_light = [{"Rs": 1, "amp": 1, "center_x": 0, "center_y": 0}]
        kwargs_anisotropy = {"r_ani": 1}

        assert kinematicAPI._copy_centers({}, kwargs_lens) == {
            "center_x": 0,
            "center_y": 0,
        }

        kinematicAPI._analytic_kinematics = False
        assert kinematicAPI._copy_centers([{}], kwargs_lens_light) == [
            {"center_x": 0, "center_y": 0}
        ]


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            # self._kwargs_aperture_kin["aperture_type"] != "IFU_grid":
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {
                "lens_model_list": ["SIS"],
                "lens_light_model_list": ["HERNQUIST"],
            }

            kwargs_aperture = {
                "aperture_type": "slit",
                "length": 1,
                "width": 1,
            }
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={"psf_type": "GAUSSIAN", "fwhm": 0.7},
                kwargs_aperture=kwargs_aperture,
                anisotropy_model="OM",
            )
            kinematicAPI.velocity_dispersion_map(
                [{"theta_E": 1, "center_x": 0, "center_y": 0}],
                [{"Rs": 1, "amp": 1, "center_x": 0, "center_y": 0}],
                {"r_ani": 1},
                direct_convolve=True,
            )

        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {"lens_light_model_list": ["HERNQUIST"]}
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kwargs_light = [{"Rs": 1, "amp": 1, "center_x": 0, "center_y": 0}]
            kinematicAPI.kinematic_light_profile(
                kwargs_light,
                MGE_fit=False,
                Hernquist_approx=True,
                r_eff=None,
                model_kinematics_bool=[True],
            )
        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {"lens_light_model_list": ["HERNQUIST"]}
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kwargs_light = [{"Rs": 1, "amp": 1, "center_x": 0, "center_y": 0}]
            kinematicAPI.kinematic_light_profile(
                kwargs_light,
                MGE_fit=False,
                Hernquist_approx=False,
                r_eff=None,
                analytic_kinematics=True,
            )
        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {
                "lens_light_model_list": ["HERNQUIST"],
                "lens_model_list": [],
            }
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kwargs_light = [{"Rs": 1, "amp": 1, "center_x": 0, "center_y": 0}]
            kinematicAPI.kinematic_lens_profiles(
                kwargs_light, MGE_fit=True, model_kinematics_bool=[True]
            )
        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {
                "lens_light_model_list": ["HERNQUIST"],
                "lens_model_list": [],
            }
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kinematicAPI.kinematic_lens_profiles(
                kwargs_lens=None, analytic_kinematics=True
            )

        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {
                "lens_light_model_list": ["HERNQUIST"],
                "lens_model_list": [],
            }
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kwargs_lens_light = [{"Rs": 1, "center_x": 0, "center_y": 0}]
            kinematicAPI.kinematic_light_profile(
                kwargs_lens_light,
                r_eff=None,
                MGE_fit=True,
                model_kinematics_bool=None,
                Hernquist_approx=False,
                kwargs_mge=None,
            )
        with self.assertRaises(ValueError):
            z_lens = 0.5
            z_source = 1.5
            kwargs_model = {
                "lens_light_model_list": ["HERNQUIST"],
                "lens_model_list": ["SIS"],
            }
            kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
            kinematicAPI = KinematicsAPI(
                z_lens,
                z_source,
                kwargs_model,
                kwargs_seeing={},
                kwargs_aperture={},
                anisotropy_model="OM",
            )
            kinematicAPI.kinematic_lens_profiles(
                kwargs_lens,
                MGE_fit=True,
                model_kinematics_bool=None,
                theta_E=None,
                kwargs_mge={},
            )

    def test_dispersion_map_grid_convolved_numeric_vs_analytical(self):
        """Test numerical vs analytical computation of IFU_grid velocity dispersion."""
        r_eff = 1.85
        theta_e = 1.63
        gamma = 2
        a_ani = 1

        def get_v_rms(
            theta_e, gamma, r_eff, a_ani=1, z_d=0.295, z_s=0.657, analytic=False
        ):
            """Compute v_rms for power-law mass and Hernquist light using Galkin's
            numerical approach.

            :param hernquist_mass: if mass in M_sun provided, uses Hernquist mass
                  profile. For debugging purpose.
            :param do_mge: True will use lenstronomy's own MGE implementation
            """
            cosmo = FlatLambdaCDM(H0=70, Om0=0.3)

            D_d = cosmo.angular_diameter_distance(z_d).value
            D_s = cosmo.angular_diameter_distance(z_s).value
            D_ds = cosmo.angular_diameter_distance_z1z2(0.5, 2.0).value

            kwargs_cosmo = {"d_d": D_d, "d_s": D_s, "d_ds": D_ds}

            xs, ys = np.meshgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20))

            kwargs_aperture = {
                "aperture_type": "IFU_grid",
                "x_grid": xs,
                "y_grid": ys,
            }

            kwargs_seeing = {
                "psf_type": "GAUSSIAN",
                "fwhm": 0.7,
            }

            kwargs_galkin_numerics = {  # 'sampling_number': 1000,
                "interpol_grid_num": 2000,
                "log_integration": True,
                "max_integrate": 100,
                "min_integrate": 0.001,
            }

            kwargs_model = {
                "lens_model_list": ["EPL"],
                "lens_light_model_list": ["HERNQUIST"],
            }

            kinematics_api = KinematicsAPI(
                z_lens=z_d,
                z_source=z_s,
                kwargs_model=kwargs_model,
                kwargs_aperture=kwargs_aperture,
                kwargs_seeing=kwargs_seeing,
                anisotropy_model="OM",
                cosmo=cosmo,
                multi_observations=False,
                # kwargs_numerics_galkin=kwargs_galkin_numerics,
                analytic_kinematics=analytic,
                Hernquist_approx=False,
                MGE_light=False,
                MGE_mass=False,  # self._cgd,
                kwargs_mge_light=None,
                kwargs_mge_mass=None,
                sampling_number=1000,
                num_kin_sampling=2000,
                num_psf_sampling=500,
            )

            kwargs_mass = [
                {
                    "theta_E": theta_e,
                    "gamma": gamma,
                    "center_x": 0,
                    "center_y": 0,
                    "e1": 0,
                    "e2": 0,
                }
            ]

            kwargs_light = [
                {"Rs": 0.551 * r_eff, "amp": 1.0, "center_x": 0, "center_y": 0}
            ]

            kwargs_anisotropy = {"r_ani": a_ani * r_eff}

            vel_dis = kinematics_api.velocity_dispersion_map(
                kwargs_mass,
                kwargs_light,
                kwargs_anisotropy,
                r_eff=r_eff,
                theta_E=theta_e,
                gamma=gamma,
                kappa_ext=0,
                direct_convolve=True,
                supersampling_factor=5,
                voronoi_bins=None,
            )

            return vel_dis

        analytic_sigma = get_v_rms(theta_e, gamma, r_eff, analytic=True)
        numeric_sigma = get_v_rms(theta_e, gamma, r_eff, analytic=False)

        # check if values match within 1%
        npt.assert_array_less(
            (analytic_sigma - numeric_sigma) / analytic_sigma,
            0.01 * np.ones_like(analytic_sigma),
        )


if __name__ == "__main__":
    pytest.main()



from lenstronomy.Util.util import convert_bool_list

__all__ = ["ProfileListBase"]


_SUPPORTED_MODELS = [
    "ARC_PERT",
    "BLANK_PLANE",
    "CHAMELEON",
    "CNFW",
    "CNFW_ELLIPSE_POTENTIAL",
    "CONST_MAG",
    "CONVERGENCE",
    "coreBURKERT",
    "CORED_DENSITY",
    "CORED_DENSITY_2",
    "CORED_DENSITY_2_MST",
    "CORED_DENSITY_EXP",
    "CORED_DENSITY_EXP_MST",
    "CORED_DENSITY_MST",
    "CORED_DENSITY_ULDM_MST",
    "CSE",
    "CTNFW_GAUSS_DEC",
    "CURVED_ARC_CONST",
    "CURVED_ARC_SIS_MST",
    "CURVED_ARC_SPP",
    "CURVED_ARC_SPT",
    "CURVED_ARC_TAN_DIFF",
    "DIPOLE",
    "DOUBLE_CHAMELEON",
    "EPL",
    "EPL_BOXYDISKY_ELL",
    "EPL_BOXYDISKY",
    "EPL_MULTIPOLE_M3M4_ELL",
    "EPL_MULTIPOLE_M3M4",
    "EPL_NUMBA",
    "EPL_Q_PHI",
    "ElliSLICE",
    "FLEXION",
    "FLEXIONFG",
    "GAUSSIAN",
    "GAUSSIAN_ELLIPSE_KAPPA",
    "GAUSSIAN_ELLIPSE_POTENTIAL",
    "GAUSSIAN_POTENTIAL",
    "GNFW",
    "GNFW_ELLIPSE_GAUSS_DEC",
    "HERNQUIST",
    "HERNQUIST_ELLIPSE_POTENTIAL",
    "HERNQUIST_ELLIPSE_CSE",
    "HESSIAN",
    "INTERPOL",
    "INTERPOL_SCALED",
    "RADIAL_INTERPOL",
    "LOS",
    "LOS_MINIMAL",
    "MULTIPOLE",
    "MULTIPOLE_ELL",
    "MULTI_GAUSSIAN",
    "MULTI_GAUSSIAN_ELLIPSE_KAPPA",
    "MULTI_GAUSSIAN_ELLIPSE_POTENTIAL",
    "NFW",
    "NFW_ELLIPSE_CSE",
    "NFW_ELLIPSE_GAUSS_DEC",
    "NFW_ELLIPSE_POTENTIAL",
    "NFW_MC",
    "NFW_MC_ELLIPSE_POTENTIAL",
    "NIE",
    "NIE_POTENTIAL",
    "NIE_SIMPLE",
    "PEMD",
    "PJAFFE",
    "PJAFFE_ELLIPSE_POTENTIAL",
    "POINT_MASS",
    "PSEUDO_DPL",
    "SERSIC",
    "SERSIC_ELLIPSE_GAUSS_DEC",
    "SERSIC_ELLIPSE_KAPPA",
    "SERSIC_ELLIPSE_POTENTIAL",
    "SHAPELETS_CART",
    "SHAPELETS_POLAR",
    "SHEAR",
    "SHEAR_GAMMA_PSI",
    "SHEAR_REDUCED",
    "SHIFT",
    "SIE",
    "SIS",
    "SIS_TRUNCATED",
    "SPEMD",
    "SPEP",
    "SPL_CORE",
    "SPP",
    "SYNTHESIS",
    "TABULATED_DEFLECTIONS",
    "TNFW",
    "TNFWC",
    "TNFW_ELLIPSE_POTENTIAL",
    "TRIPLE_CHAMELEON",
    "ULDM",
]


class ProfileListBase(object):
    """Class that manages the list of lens model class instances.

    This class is applicable for single plane and multi plane lensing
    """

    def __init__(
        self,
        lens_model_list,
        numerical_alpha_class=None,
        lens_redshift_list=None,
        z_source_convention=None,
        kwargs_interp=None,
        kwargs_synthesis=None,
    ):
        """

        :param lens_model_list: list of strings with lens model names
        :param numerical_alpha_class: an instance of a custom class for use in NumericalAlpha() lens model
         deflection angles as a lens model. See the documentation in Profiles.numerical_deflections
        :param kwargs_interp: interpolation keyword arguments specifying the numerics.
         See description in the Interpolate() class. Only applicable for 'INTERPOL' and 'INTERPOL_SCALED' models.
        :param kwargs_synthesis: keyword arguments for the 'SYNTHESIS' lens model, if applicable
        """
        self.func_list = self._load_model_instances(
            lens_model_list,
            custom_class=numerical_alpha_class,
            lens_redshift_list=lens_redshift_list,
            z_source_convention=z_source_convention,
            kwargs_interp=kwargs_interp,
            kwargs_synthesis=kwargs_synthesis,
        )
        self._num_func = len(self.func_list)
        self._model_list = lens_model_list

    def _load_model_instances(
        self,
        lens_model_list,
        custom_class=None,
        lens_redshift_list=None,
        z_source_convention=None,
        kwargs_interp=None,
        kwargs_synthesis=None,
    ):
        if lens_redshift_list is None:
            lens_redshift_list = [None] * len(lens_model_list)
        if kwargs_interp is None:
            kwargs_interp = {}
        if kwargs_synthesis is None:
            kwargs_synthesis = {}
        func_list = []
        imported_classes = {}
        for i, lens_type in enumerate(lens_model_list):
            # those models require a new instance per profile as some pre-computations are different when parameters or
            # other settings are changed. For example, the 'INTERPOL' model needs to know the specific map to be
            # interpolated.
            if lens_type in [
                "CHAMELEON",
                "CTNFW_GAUSS_DEC",
                "DOUBLE_CHAMELEON",
                "INTERPOL",
                "INTERPOL_SCALED",
                "NFW_ELLIPSE_GAUSS_DEC",
                "NFW_MC",
                "NFW_MC_ELLIPSE_POTENTIAL",
                "NIE",
                "NIE_SIMPLE",
                "RADIAL_INTERPOL",
                "TRIPLE_CHAMELEON",
                "EPL",
            ]:
                lensmodel_class = lens_class(
                    lens_type,
                    custom_class=custom_class,
                    z_lens=lens_redshift_list[i],
                    z_source=z_source_convention,
                    kwargs_interp=kwargs_interp,
                    kwargs_synthesis=kwargs_synthesis,
                )
            else:
                if lens_type not in imported_classes.keys():
                    lensmodel_class = lens_class(
                        lens_type,
                        custom_class=custom_class,
                        kwargs_interp=kwargs_interp,
                        kwargs_synthesis=kwargs_synthesis,
                    )
                    imported_classes.update({lens_type: lensmodel_class})
                else:
                    lensmodel_class = imported_classes[lens_type]
            func_list.append(lensmodel_class)
        return func_list

    def _bool_list(self, k=None):
        """Returns a bool list of the length of the lens models if k = None: returns
        bool list with True's if k is int, returns bool list with False's but k'th is
        True if k is a list of int, e.g. [0, 3, 5], returns a bool list with True's in
        the integers listed and False elsewhere if k is a boolean list, checks for size
        to match the numbers of models and returns it.

        :param k: None, int, or list of ints
        :return: bool list
        """
        return convert_bool_list(n=self._num_func, k=k)

    def set_static(self, kwargs_list):
        """

        :param kwargs_list: list of keyword arguments for each profile
        :return: kwargs_list
        """
        for i, func in enumerate(self.func_list):
            func.set_static(**kwargs_list[i])
        return kwargs_list

    def set_dynamic(self):
        """Frees cache set by static model (if exists) and re-computes all lensing
        quantities each time a definition is called assuming different parameters are
        executed. This is the default mode if not specified as set_static()

        :return: None
        """
        for i, func in enumerate(self.func_list):
            func.set_dynamic()

    def model_info(self):
        """Shows what models are being initialized and what parameters are being
        requested for.

        :return: None
        """
        for i, func in enumerate(self.func_list):
            print(
                "Lens model %s is %s with parameters %s"
                % (i, self._model_list[i], func.param_names)
            )


def lens_class(
    lens_type,
    custom_class=None,
    kwargs_interp=None,
    kwargs_synthesis=None,
    z_lens=None,
    z_source=None,
):
    """Generate class instance of single lens.

    :param lens_type: string, lens model type
    :param custom_class: custom class
    :param z_lens: lens redshift # currently only used in NFW_MC model as this is
        redshift dependent
    :param z_source: source redshift # currently only used in NFW_MC model as this is
        redshift dependent
    :param kwargs_interp: interpolation keyword arguments specifying the numerics. See
        description in the Interpolate() class. Only applicable for 'INTERPOL' and
        'INTERPOL_SCALED' models.
    :return: class instance of the lens model type
    """

    if lens_type == "ARC_PERT":
        from lenstronomy.LensModel.Profiles.arc_perturbations import (
            ArcPerturbations,
        )

        return ArcPerturbations()
    if lens_type == "BLANK_PLANE":
        from lenstronomy.LensModel.Profiles.blank_plane import BlankPlane

        return BlankPlane()
    elif lens_type == "CHAMELEON":
        from lenstronomy.LensModel.Profiles.chameleon import Chameleon

        return Chameleon()
    elif lens_type == "CNFW":
        from lenstronomy.LensModel.Profiles.cnfw import CNFW

        return CNFW()
    elif lens_type == "CNFW_ELLIPSE_POTENTIAL":
        from lenstronomy.LensModel.Profiles.cnfw_ellipse_potential import (
            CNFWEllipsePotential,
        )

        return CNFWEllipsePotential()
    elif lens_type == "CONST_MAG":
        from lenstronomy.LensModel.Profiles.const_mag import ConstMag

        return ConstMag()
    elif lens_type == "CONVERGENCE":
        from lenstronomy.LensModel.Profiles.convergence import Convergence

        return Convergence()
    elif lens_type == "coreBURKERT":
        from lenstronomy.LensModel.Profiles.coreBurkert import CoreBurkert

        return CoreBurkert()
    elif lens_type == "CORED_DENSITY":
        from lenstronomy.LensModel.Profiles.cored_density import CoredDensity

        return CoredDensity()
    elif lens_type == "CORED_DENSITY_2":
        from lenstronomy.LensModel.Profiles.cored_density_2 import CoredDensity2

        return CoredDensity2()
    elif lens_type == "CORED_DENSITY_2_MST":
        from lenstronomy.LensModel.Profiles.cored_density_mst import CoredDensityMST

        return CoredDensityMST(profile_type="CORED_DENSITY_2")
    elif lens_type == "CORED_DENSITY_EXP":
        from lenstronomy.LensModel.Profiles.cored_density_exp import CoredDensityExp

        return CoredDensityExp()
    elif lens_type == "CORED_DENSITY_EXP_MST":
        from lenstronomy.LensModel.Profiles.cored_density_mst import CoredDensityMST

        return CoredDensityMST(profile_type="CORED_DENSITY_EXP")
    elif lens_type == "CORED_DENSITY_MST":
        from lenstronomy.LensModel.Profiles.cored_density_mst import CoredDensityMST

        return CoredDensityMST(profile_type="CORED_DENSITY")
    elif lens_type == "CORED_DENSITY_ULDM_MST":
        from lenstronomy.LensModel.Profiles.cored_density_mst import CoredDensityMST

        return CoredDensityMST(profile_type="CORED_DENSITY_ULDM")
    elif lens_type == "CSE":
        from lenstronomy.LensModel.Profiles.cored_steep_ellipsoid import CSE

        return CSE()
    elif lens_type == "CTNFW_GAUSS_DEC":
        from lenstronomy.LensModel.Profiles.gauss_decomposition import CTNFWGaussDec

        return CTNFWGaussDec()
    elif lens_type == "CURVED_ARC_CONST":
        from lenstronomy.LensModel.Profiles.curved_arc_const import CurvedArcConst

        return CurvedArcConst()
    elif lens_type == "CURVED_ARC_CONST_MST":
        from lenstronomy.LensModel.Profiles.curved_arc_const import CurvedArcConstMST

        return CurvedArcConstMST()
    elif lens_type == "CURVED_ARC_SIS_MST":
        from lenstronomy.LensModel.Profiles.curved_arc_sis_mst import CurvedArcSISMST

        return CurvedArcSISMST()
    elif lens_type == "CURVED_ARC_SPP":
        from lenstronomy.LensModel.Profiles.curved_arc_spp import CurvedArcSPP

        return CurvedArcSPP()
    elif lens_type == "CURVED_ARC_SPT":
        from lenstronomy.LensModel.Profiles.curved_arc_spt import CurvedArcSPT

        return CurvedArcSPT()
    elif lens_type == "CURVED_ARC_TAN_DIFF":
        from lenstronomy.LensModel.Profiles.curved_arc_tan_diff import (
            CurvedArcTanDiff,
        )

        return CurvedArcTanDiff()
    elif lens_type == "DIPOLE":
        from lenstronomy.LensModel.Profiles.dipole import Dipole

        return Dipole()
    elif lens_type == "DOUBLE_CHAMELEON":
        from lenstronomy.LensModel.Profiles.chameleon import DoubleChameleon

        return DoubleChameleon()
    elif lens_type == "EPL":
        from lenstronomy.LensModel.Profiles.epl import EPL

        return EPL()
    elif lens_type == "EPL_BOXYDISKY_ELL":
        from lenstronomy.LensModel.Profiles.epl_boxydisky import EPL_BOXYDISKY_ELL

        return EPL_BOXYDISKY_ELL()
    elif lens_type == "EPL_BOXYDISKY":
        from lenstronomy.LensModel.Profiles.epl_boxydisky import EPL_BOXYDISKY

        return EPL_BOXYDISKY()
    elif lens_type == "EPL_MULTIPOLE_M3M4_ELL":
        from lenstronomy.LensModel.Profiles.epl_multipole_m3m4 import (
            EPL_MULTIPOLE_M3M4_ELL,
        )

        return EPL_MULTIPOLE_M3M4_ELL()
    elif lens_type == "EPL_MULTIPOLE_M3M4":
        from lenstronomy.LensModel.Profiles.epl_multipole_m3m4 import (
            EPL_MULTIPOLE_M3M4,
        )

        return EPL_MULTIPOLE_M3M4()
    elif lens_type == "EPL_NUMBA":
        from lenstronomy.LensModel.Profiles.epl_numba import EPL_numba

        return EPL_numba()
    elif lens_type == "EPL_Q_PHI":
        from lenstronomy.LensModel.Profiles.epl import EPLQPhi

        return EPLQPhi()
    elif lens_type == "ElliSLICE":
        from lenstronomy.LensModel.Profiles.elliptical_density_slice import (
            ElliSLICE,
        )

        return ElliSLICE()
    elif lens_type == "FLEXION":
        from lenstronomy.LensModel.Profiles.flexion import Flexion

        return Flexion()
    elif lens_type == "FLEXIONFG":
        from lenstronomy.LensModel.Profiles.flexionfg import Flexionfg

        return Flexionfg()
    elif lens_type == "GAUSSIAN":
        from lenstronomy.LensModel.Profiles.gaussian import Gaussian

        return Gaussian()
    elif lens_type == "GAUSSIAN_ELLIPSE_KAPPA":
        from lenstronomy.LensModel.Profiles.gaussian_ellipse_kappa import (
            GaussianEllipseKappa,
        )

        return GaussianEllipseKappa()
    elif lens_type == "GAUSSIAN_ELLIPSE_POTENTIAL":
        from lenstronomy.LensModel.Profiles.gaussian_ellipse_potential import (
            GaussianEllipsePotential,
        )

        return GaussianEllipsePotential()
    elif lens_type == "GAUSSIAN_POTENTIAL":
        from lenstronomy.LensModel.Profiles.gaussian_potential import GaussianPotential

        return GaussianPotential()
    elif lens_type == "GNFW":
        from lenstronomy.LensModel.Profiles.gnfw import GNFW

        return GNFW()
    elif lens_type == "GNFW_ELLIPSE_GAUSS_DEC":
        from lenstronomy.LensModel.Profiles.gauss_decomposition import (
            GeneralizedNFWEllipseGaussDec,
        )

        return GeneralizedNFWEllipseGaussDec()
    elif lens_type == "HERNQUIST":
        from lenstronomy.LensModel.Profiles.hernquist import Hernquist

        return Hernquist()
    elif lens_type == "HERNQUIST_ELLIPSE_POTENTIAL":
        from lenstronomy.LensModel.Profiles.hernquist_ellipse_potential import (
            HernquistEllipsePotential,
        )

        return HernquistEllipsePotential()
    elif lens_type == "HERNQUIST_ELLIPSE_CSE":
        from lenstronomy.LensModel.Profiles.hernquist_ellipse_cse import (
            HernquistEllipseCSE,
        )

        return HernquistEllipseCSE()
    elif lens_type == "HESSIAN":
        from lenstronomy.LensModel.Profiles.hessian import Hessian

        return Hessian()
    elif lens_type == "INTERPOL":
        from lenstronomy.LensModel.Profiles.interpol import Interpol

        return Interpol(**kwargs_interp)
    elif lens_type == "INTERPOL_SCALED":
        from lenstronomy.LensModel.Profiles.interpol import InterpolScaled

        return InterpolScaled(**kwargs_interp)
    elif lens_type == "LOS":
        from lenstronomy.LensModel.LineOfSight.LOSModels.los import LOS

        return LOS()
    elif lens_type == "LOS_MINIMAL":
        from lenstronomy.LensModel.LineOfSight.LOSModels.los_minimal import (
            LOSMinimal,
        )

        return LOSMinimal()
    elif lens_type == "MULTIPOLE":
        from lenstronomy.LensModel.Profiles.multipole import Multipole

        return Multipole()
    elif lens_type == "MULTIPOLE_ELL":
        from lenstronomy.LensModel.Profiles.multipole import EllipticalMultipole

        return EllipticalMultipole()
    elif lens_type == "MULTI_GAUSSIAN":
        from lenstronomy.LensModel.Profiles.multi_gaussian import (
            MultiGaussian,
        )

        return MultiGaussian()
    elif lens_type == "MULTI_GAUSSIAN_ELLIPSE_KAPPA":
        from lenstronomy.LensModel.Profiles.multi_gaussian_ellipse_kappa import (
            MultiGaussianEllipseKappa,
        )

        return MultiGaussianEllipseKappa()
    elif lens_type == "MULTI_GAUSSIAN_ELLIPSE_POTENTIAL":
        from lenstronomy.LensModel.Profiles.multi_gaussian import (
            MultiGaussianEllipsePotential,
        )

        return MultiGaussianEllipsePotential()
    elif lens_type == "NFW":
        from lenstronomy.LensModel.Profiles.nfw import NFW

        return NFW()
    elif lens_type == "NFW_ELLIPSE_POTENTIAL":
        from lenstronomy.LensModel.Profiles.nfw_ellipse_potential import (
            NFWEllipsePotential,
        )

        return NFWEllipsePotential()
    elif lens_type == "NFW_ELLIPSE_CSE":
        from lenstronomy.LensModel.Profiles.nfw_ellipse_cse import NFW_ELLIPSE_CSE

        return NFW_ELLIPSE_CSE()
    elif lens_type == "NFW_ELLIPSE_GAUSS_DEC":
        from lenstronomy.LensModel.Profiles.gauss_decomposition import (
            NFWEllipseGaussDec,
        )

        return NFWEllipseGaussDec()
    elif lens_type == "NFW_MC":
        from lenstronomy.LensModel.Profiles.nfw_mass_concentration import NFWMC

        return NFWMC(z_lens=z_lens, z_source=z_source)
    elif lens_type == "NFW_MC_ELLIPSE_POTENTIAL":
        from lenstronomy.LensModel.Profiles.nfw_mass_concentration_ellipse import (
            NFWMCEllipsePotential,
        )

        return NFWMCEllipsePotential(z_lens=z_lens, z_source=z_source)
    elif lens_type == "NIE":
        from lenstronomy.LensModel.Profiles.nie import NIE

        return NIE()
    elif lens_type == "NIE_POTENTIAL":
        from lenstronomy.LensModel.Profiles.nie_potential import NIE_POTENTIAL

        return NIE_POTENTIAL()
    elif lens_type == "NIE_SIMPLE":
        from lenstronomy.LensModel.Profiles.nie import NIEMajorAxis

        return NIEMajorAxis()
    elif lens_type == "PEMD":
        from lenstronomy.LensModel.Profiles.pemd import PEMD

        return PEMD()
    elif lens_type == "PJAFFE":
        from lenstronomy.LensModel.Profiles.pseudo_jaffe import PseudoJaffe

        return PseudoJaffe()
    elif lens_type == "PJAFFE_ELLIPSE_POTENTIAL":
        from lenstronomy.LensModel.Profiles.pseudo_jaffe_ellipse_potential import (
            PseudoJaffeEllipsePotential,
        )

        return PseudoJaffeEllipsePotential()
    elif lens_type == "POINT_MASS":
        from lenstronomy.LensModel.Profiles.point_mass import PointMass

        return PointMass()
    elif lens_type == "PSEUDO_DPL":
        from lenstronomy.LensModel.Profiles.pseudo_double_powerlaw import (
            PseudoDoublePowerlaw,
        )

        return PseudoDoublePowerlaw()
    elif lens_type == "RADIAL_INTERPOL":
        from lenstronomy.LensModel.Profiles.radial_interpolated import (
            RadialInterpolate,
        )

        return RadialInterpolate()
    elif lens_type == "SERSIC":
        from lenstronomy.LensModel.Profiles.sersic import Sersic

        return Sersic()
    elif lens_type == "SERSIC_ELLIPSE_GAUSS_DEC":
        from lenstronomy.LensModel.Profiles.gauss_decomposition import (
            SersicEllipseGaussDec,
        )

        return SersicEllipseGaussDec()
    elif lens_type == "SERSIC_ELLIPSE_KAPPA":
        from lenstronomy.LensModel.Profiles.sersic_ellipse_kappa import (
            SersicEllipseKappa,
        )

        return SersicEllipseKappa()
    elif lens_type == "SERSIC_ELLIPSE_POTENTIAL":
        from lenstronomy.LensModel.Profiles.sersic_ellipse_potential import (
            SersicEllipsePotential,
        )

        return SersicEllipsePotential()
    elif lens_type == "SHAPELETS_CART":
        from lenstronomy.LensModel.Profiles.shapelet_pot_cartesian import (
            CartShapelets,
        )

        return CartShapelets()
    elif lens_type == "SHAPELETS_POLAR":
        from lenstronomy.LensModel.Profiles.shapelet_pot_polar import PolarShapelets

        return PolarShapelets()
    elif lens_type == "SHIFT":
        from lenstronomy.LensModel.Profiles.constant_shift import Shift

        return Shift()
    elif lens_type == "SHEAR":
        from lenstronomy.LensModel.Profiles.shear import Shear

        return Shear()
    elif lens_type == "SHEAR_GAMMA_PSI":
        from lenstronomy.LensModel.Profiles.shear import ShearGammaPsi

        return ShearGammaPsi()
    elif lens_type == "SHEAR_REDUCED":
        from lenstronomy.LensModel.Profiles.shear import ShearReduced

        return ShearReduced()
    elif lens_type == "SIE":
        from lenstronomy.LensModel.Profiles.sie import SIE

        return SIE()
    elif lens_type == "SIS":
        from lenstronomy.LensModel.Profiles.sis import SIS

        return SIS()
    elif lens_type == "SIS_TRUNCATED":
        from lenstronomy.LensModel.Profiles.sis_truncate import SIS_truncate

        return SIS_truncate()
    elif lens_type == "SPEMD":
        from lenstronomy.LensModel.Profiles.spemd import SPEMD

        return SPEMD()
    elif lens_type == "SPEP":
        from lenstronomy.LensModel.Profiles.spep import SPEP

        return SPEP()
    elif lens_type == "SPL_CORE":
        from lenstronomy.LensModel.Profiles.splcore import SPLCORE

        return SPLCORE()
    elif lens_type == "SPP":
        from lenstronomy.LensModel.Profiles.spp import SPP

        return SPP()
    elif lens_type == "SYNTHESIS":
        from lenstronomy.LensModel.Profiles.synthesis import SynthesisProfile

        return SynthesisProfile(**kwargs_synthesis)
    elif lens_type == "TABULATED_DEFLECTIONS":
        from lenstronomy.LensModel.Profiles.numerical_deflections import (
            TabulatedDeflections,
        )

        return TabulatedDeflections(custom_class)
    elif lens_type == "TNFW":
        from lenstronomy.LensModel.Profiles.tnfw import TNFW

        return TNFW()
    elif lens_type == "TNFWC":
        from lenstronomy.LensModel.Profiles.nfw_core_truncated import TNFWC

        return TNFWC()
    elif lens_type == "TNFW_ELLIPSE_POTENTIAL":
        from lenstronomy.LensModel.Profiles.tnfw_ellipse_potential import (
            TNFWELLIPSEPotential,
        )

        return TNFWELLIPSEPotential()
    elif lens_type == "TRIPLE_CHAMELEON":
        from lenstronomy.LensModel.Profiles.chameleon import TripleChameleon

        return TripleChameleon()
    elif lens_type == "ULDM":
        from lenstronomy.LensModel.Profiles.uldm import Uldm

        return Uldm()
    # when adding a new profile, insert the corresponding elif statement in its
    # alphabetical position

    else:
        raise ValueError(
            "%s is not a valid lens model. Supported are: %s."
            % (lens_type, _SUPPORTED_MODELS)
        )



import lenstronomy.Util.param_util as param_util
import numpy as np

__all__ = ["Hernquist", "HernquistEllipse"]


class Hernquist(object):
    """Class for Hernquist lens light (2d projected light/mass distribution)."""

    def __init__(self):
        from lenstronomy.LensModel.Profiles.hernquist import Hernquist as Hernquist_lens

        self.lens = Hernquist_lens()
        self.param_names = ["amp", "Rs", "center_x", "center_y"]
        self.lower_limit_default = {
            "amp": 0,
            "Rs": 0,
            "center_x": -100,
            "center_y": -100,
        }
        self.upper_limit_default = {
            "amp": 100,
            "Rs": 100,
            "center_x": 100,
            "center_y": 100,
        }

    def function(self, x, y, amp, Rs, center_x=0, center_y=0):
        """

        :param x: x-position
        :param y: y-position
        :param amp: surface brightness amplitude
        :param Rs: scale radius: half-light radius = Rs / 0.551
        :param center_x: centroid in x-direction
        :param center_y: centroid in y-direction
        :return: surface brightness
        """
        rho0 = self.lens.sigma2rho(amp, Rs)
        return self.lens.density_2d(x, y, rho0, Rs, center_x, center_y)

    def light_3d(self, r, amp, Rs):
        """

        :param r: 3d radius (in angular units)
        :param amp: surface brightness amplitude
        :param Rs: scale radius: half-light radius = Rs / 0.551
        :return:
        """
        rho0 = self.lens.sigma2rho(amp, Rs)
        return self.lens.density(r, rho0, Rs)

    @staticmethod
    def total_flux(amp, Rs):
        """

        :param amp: surface brightness amplitude
        :param Rs: scale radius: half-light radius = Rs / 0.551
        :return: total integrated flux of profile
        """
        rhos = amp / Rs
        m_tot = 2 * np.pi * rhos * Rs**3
        return m_tot


class HernquistEllipse(object):
    """Class for elliptical pseudo Jaffe lens light (2d projected light/mass
    distribution."""

    param_names = ["amp", "Rs", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "Rs": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "Rs": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        from lenstronomy.LensModel.Profiles.hernquist import Hernquist as Hernquist_lens

        self.lens = Hernquist_lens()
        self.spherical = Hernquist()

    def function(self, x, y, amp, Rs, e1, e2, center_x=0, center_y=0):
        """

        :param x: x-position
        :param y: y-position
        :param amp: surface brightness amplitude
        :param Rs: scale radius: half-light radius = Rs / 0.551
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param center_x: centroid in x-direction
        :param center_y: centroid in y-direction
        :return: surface brightness
        """
        x_, y_ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x, center_y
        )
        return self.spherical.function(x_, y_, amp, Rs)

    def light_3d(self, r, amp, Rs, e1=0, e2=0):
        """

        :param r: 3d radius (in angular units)
        :param amp: surface brightness amplitude
        :param Rs: scale radius: half-light radius = Rs / 0.551
        :param e1: eccentricity component
        :param e2: eccentricity component
        :return: flux density in 3d
        """
        rho0 = self.lens.sigma2rho(amp, Rs)
        return self.lens.density(r, rho0, Rs)

    def total_flux(self, amp, Rs, e1=0, e2=0):
        """

        :param amp: surface brightness amplitude
        :param Rs: scale radius: half-light radius = Rs / 0.551
        :param e1: eccentricity component
        :param e2: eccentricity component
        :return: total integrated flux
        """
        return self.spherical.total_flux(amp=amp, Rs=Rs)



import numpy as np
from lenstronomy.Util import class_creator


class TracerLikelihood(object):
    """Class to evaluate the tracer map."""

    def __init__(
        self,
        tracer_data,
        kwargs_model,
        kwargs_imaging,
        tracer_likelihood_mask=None,
        linear_solver=True,
    ):
        """

        :param tracer_bands_compute: list of bools with same length as data objects, indicates which "band" to include in the
         fitting
        :param tracer_likelihood_mask: boolean 2d arrays of size of images marking the pixels to be
        :param linear_solver: boolean, if True, linear solver for surface brigthness amplitude is activated
         evaluated in the likelihood
        """
        self._tracer_light_model_band = kwargs_model.get("tracer_source_band", 0)
        self.image_model = class_creator.create_im_sim(
            kwargs_model=kwargs_model, linear_solver=linear_solver, **kwargs_imaging
        )
        self.tracerModel = class_creator.create_tracer_model(
            tracer_data, kwargs_model, tracer_likelihood_mask=tracer_likelihood_mask
        )

    def logL(
        self,
        kwargs_tracer_source,
        kwargs_lens=None,
        kwargs_source=None,
        kwargs_lens_light=None,
        kwargs_ps=None,
        kwargs_special=None,
        kwargs_extinction=None,
        param=None,
    ):
        """

        :param kwargs_lens: lens model keyword argument list according to LensModel module
        :param kwargs_source: source light keyword argument list according to LightModel module
        :param kwargs_lens_light: deflector light (not lensed) keyword argument list according to LightModel module
        :param kwargs_ps: point source keyword argument list according to PointSource module
        :param kwargs_special: special keyword argument list as part of the Param module
        :param kwargs_extinction: extinction parameter keyword argument list according to LightModel module
        :return: log likelihood of the data given the model
        """
        (
            kwargs_lens,
            kwargs_source,
            kwargs_lens_light,
            kwargs_ps,
        ) = self.image_model.update_linear_kwargs(
            param,
            model_band=self._tracer_light_model_band,
            kwargs_lens=kwargs_lens,
            kwargs_source=kwargs_source,
            kwargs_lens_light=kwargs_lens_light,
            kwargs_ps=kwargs_ps,
        )

        logL = self.tracerModel.likelihood_data_given_model(
            kwargs_tracer_source=kwargs_tracer_source,
            kwargs_lens=kwargs_lens,
            kwargs_source=kwargs_source,
            kwargs_extinction=kwargs_extinction,
            kwargs_special=kwargs_special,
        )
        if np.isnan(logL) is True:
            return -(10**15)
        return logL

    @property
    def num_data(self):
        """

        :return: number of image data points
        """
        return self.tracerModel.num_data_evaluate

    def reset_point_source_cache(self, cache=True):
        """

        :param cache: boolean
        :return: None
        """
        self.tracerModel.reset_point_source_cache(cache=cache)



import numpy as np
import copy

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def bkg_noise(
    readout_noise, exposure_time, sky_brightness, pixel_scale, num_exposures=1
):
    """Computes the expected Gaussian background noise of a pixel in units of
    counts/second.

    :param readout_noise: noise added per readout
    :param exposure_time: exposure time per exposure (in seconds)
    :param sky_brightness: counts per second per unit arcseconds square
    :param pixel_scale: size of pixel in units arcseonds
    :param num_exposures: number of exposures (with same exposure time) to be co-added
    :return: estimated Gaussian noise sqrt(variance)
    """
    exposure_time_tot = num_exposures * exposure_time
    readout_noise_tot = num_exposures * readout_noise**2  # square of readout noise
    sky_per_pixel = sky_brightness * pixel_scale**2
    sky_brightness_tot = exposure_time_tot * sky_per_pixel
    sigma_bkg = np.sqrt(readout_noise_tot + sky_brightness_tot) / exposure_time_tot
    return sigma_bkg


@export
def flux_noise(cps_pixel, exposure_time):
    """Computes the variance of the shot noise Gaussian approximation of Poisson noise
    term.

    :param cps_pixel: counts per second of the intensity per pixel unit
    :param exposure_time: total exposure time (in units seconds or equivalent unit as
        cps_pixel)
    :return: sqrt(variance) of pixel value
    """
    return cps_pixel / np.sqrt(exposure_time)


@export
def magnitude2cps(magnitude, magnitude_zero_point):
    """Converts an apparent magnitude to counts per second.

    The zero point of an instrument, by definition, is the magnitude of an object that produces one count
    (or data number, DN) per second. The magnitude of an arbitrary object producing DN counts in an observation of
    length EXPTIME is therefore:
    m = -2.5 x log10(DN / EXPTIME) + ZEROPOINT

    :param magnitude: astronomical magnitude
    :param magnitude_zero_point: magnitude zero point (astronomical magnitude with 1 count per second)
    :return: counts per second of astronomical object
    """
    delta_m = magnitude - magnitude_zero_point
    counts = 10 ** (-delta_m / 2.5)
    return counts


@export
def cps2magnitude(cps, magnitude_zero_point):
    """

    :param cps: float, count-per-second
    :param magnitude_zero_point: magnitude zero point
    :return: magnitude for given counts
    """
    delta_m = -np.log10(cps) * 2.5
    magnitude = delta_m + magnitude_zero_point
    return magnitude


@export
def absolute2apparent_magnitude(absolute_magnitude, d_parsec):
    """Converts absolute to apparent magnitudes.

    :param absolute_magnitude: absolute magnitude of object
    :param d_parsec: distance to object in units parsec
    :return: apparent magnitude
    """
    m_apparent = 5.8 * (np.log10(d_parsec) - 1) + absolute_magnitude
    return m_apparent


@export
def adu2electrons(adu, ccd_gain):
    """Converts analog-to-digital units into electron counts.

    :param adu: counts in analog-to-digital unit
    :param ccd_gain: CCD gain, meaning how many electrons are counted per unit ADU
    :return: counts in electrons
    """
    return adu * ccd_gain


@export
def electrons2adu(electrons, ccd_gain):
    """Converts electron counts into analog-to-digital unit.

    :param electrons: number of electrons received on detector
    :param ccd_gain: CCD gain, meaning how many electrons are counted per unit ADU
    :return: adu value in Analog-to-digital units corresponding to electron count
    """
    return electrons / ccd_gain


def magnitude2amplitude(light_model_class, kwargs_light_mag, magnitude_zero_point):
    """Translates astronomical magnitudes to lenstronomy linear 'amp' parameters for
    LightModel objects.

    :param light_model_class: LightModel() class instance
    :param kwargs_light_mag: list of light model parameter dictionary with 'magnitude'
        instead of 'amp'
    :param magnitude_zero_point: magnitude zero point
    :return: list of light model parameter dictionary with 'amp'
    """
    kwargs_light_amp = copy.deepcopy(kwargs_light_mag)
    if kwargs_light_mag is not None:
        for i, kwargs_mag in enumerate(kwargs_light_mag):
            kwargs_new = kwargs_light_amp[i]
            del kwargs_new["magnitude"]
            cps_norm = light_model_class.total_flux(
                kwargs_list=kwargs_light_amp, norm=True, k=i
            )[0]
            magnitude = kwargs_mag["magnitude"]
            cps = magnitude2cps(magnitude, magnitude_zero_point=magnitude_zero_point)
            amp = cps / cps_norm
            kwargs_new["amp"] = amp
    return kwargs_light_amp
